"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .. import Container

###############################################################################
class Layer(Container):						# pylint: disable=abstract-method
	""" Base class for layers, which are containers that produce
		backend-specific layers, or which wrap backend-specific layers with
		additional backend-specific operations.
	"""

	###########################################################################
	def terminal(self):
		""" The identifying feature of a Layer is that it is terminal, meaning
			that it actually produces layers.
		"""
		return True

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import pytest

from kur.loss import MeanSquaredError
from kur.optimizer import Adam
from kur.model import Trainer

@pytest.fixture
def loss():
	return MeanSquaredError()

@pytest.fixture
def optimizer():
	return Adam()

###############################################################################
class TestTrain:
	""" Tests for the Trainer
	"""

	###########################################################################
	def test_trainer_with_optimizer(self, simple_model, loss, optimizer):
		""" Tests if we can compile a Trainer instance with an optimizer.
		"""
		simple_model.parse(None)
		simple_model.build()
		trainer = Trainer(
			model=simple_model,
			loss=loss,
			optimizer=optimizer
		)
		trainer.compile()

	###########################################################################
	def test_trainer_without_optimizer(self, simple_model, loss, optimizer):
		""" Tests if we can compile a Trainer instance without an optimizer.
		"""
		simple_model.parse(None)
		simple_model.build()
		trainer = Trainer(
			model=simple_model,
			loss=loss,
			optimizer=optimizer
		)
		trainer.compile()

	###########################################################################
	def test_training(self, simple_model, loss, optimizer, simple_data):
		""" Tests that a single training epoch can succeed.
		"""
		simple_model.parse(None)
		simple_model.build()
		trainer = Trainer(
			model=simple_model,
			loss=loss,
			optimizer=optimizer
		)
		trainer.train(provider=simple_data, epochs=1)

	###########################################################################
	def test_testing(self, simple_model, loss, simple_data):
		""" Tests that a testing run can succeed.
		"""
		simple_model.parse(None)
		simple_model.build()
		trainer = Trainer(
			model=simple_model,
			loss=loss
		)
		trainer.test(provider=simple_data)

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import warnings
from . import ParsingError
from ..utils import get_subclasses

###############################################################################
class Container:
	""" The base class for all parseable entries in the model representation.

		# General container features

		- name: unique, immutable (sticks to the first named container)
		- tag: non-unique, mutable (sticks to the last named container)
		- oldest: non-unique, immutable (sticks to the first named container)
		- inputs: the layers that are used as input
		- sink
		- when:
		- skip:
	"""

	# Used for generating unique names.
	counter = {}

	###########################################################################
	@classmethod
	def get_container_name(cls):
		""" Returns the name of the container class.

			This is the name used to determine which container needs to be
			instantiated. It can be overriden in derived classes if the name of
			the class isn't the same as the key used in the specification.

			# Return value

			A lower-case string unique to this container class.
		"""
		return cls.__name__.lower()

	###########################################################################
	@classmethod
	def unique_name(cls):
		""" Generates a unique name for this instance.
		"""
		if cls in Container.counter:
			val = Container.counter[cls]
		else:
			val = 0
		Container.counter[cls] = val + 1
		return '..{}.{}'.format(cls.get_container_name(), val)

	###########################################################################
	@staticmethod
	def create_container_from_data(data, **kwargs):
		""" Factory method for creating containers.
		"""
		cls = Container.find_container_for_data(data)
		return cls(data, **kwargs)

	###########################################################################
	@staticmethod
	def find_container_for_data(data):
		""" Finds a class object corresponding to a data blob.
		"""
		for cls in get_subclasses(Container):
			if cls.get_container_name() in data:
				return cls
		raise ValueError('No such container for data: {}'.format(data))

	###########################################################################
	def __init__(self, data):
		""" Creates a new container.
		"""
		self.data = data
		self.tags = []
		self.name = None
		self.oldest = []
		self.inputs = []
		self.args = None
		self.sink = False

		self.parent = None
		self.children = []

		self._built = None
		self._parsed = False

	###########################################################################
	def __str__(self):
		""" Return a string representation.

			This uses the container's declared name in the representation.
		"""
		return '{type}(name={name})'.format(
			type=self.get_container_name(),
			name=self.name
		)

	###########################################################################
	def __repr__(self):
		""" Return a string representation.

			This uses the container's class name in the representation.
		"""
		return '{type}(name={name})'.format(
			type=self.__class__.__name__,
			name=self.name
		)

	###########################################################################
	def is_parsed(self):
		""" Returns True if this container has been parsed already.
		"""
		return self._parsed

	###########################################################################
	def add_child(self, container):
		""" Attaches an existing container as a child to this container.
		"""
		if container.parent is not None:
			raise ParsingError('Cannot add a child container which already '
				'has parents.')
		container.parent = self
		self.children.append(container)

	###########################################################################
	def remove_child(self, container, recursive=True):
		""" Removes a child container by name.

			# Arguments

			container: str or Container. If a Container instance, the container
				to remove. If a string, the name of the child container to
				remove.
			recursive: bool (default: True). If True, searches recursively for
				the child container; otherwise, only looks at this container's
				immediate children.

			# Return value

			None

			# Exceptions

			If the child container cannot be found, a ValueError exception is
			raised.
		"""
		name = container.name if isinstance(container, Container) \
			else container

		child = self.get_child_by_name(name, recursive=recursive)
		if child is None:
			raise ValueError('No such child found: {}'.format(name))
		target = child.parent
		if target is None:
			raise ValueError('Malformed children. A parent is missing for "{}"'
				.format(child.name))
		child.parent = None
		target.children.remove(child)

	###########################################################################
	def new_child_from_data(self, data):
		""" Creates a new container from data and attaches it as a child.
		"""
		child = Container.create_container_from_data(data)
		self.add_child(child)
		return child

	###########################################################################
	def get_children(self, recursive=False, include_self=False):
		""" Returns this container's children.

			# Arguments

			recursive: bool (default: False). If True, recursively searches
				each container's children for additional children.
			include_self: bool (default: False). If True, yields this container
				as well.

			# Return value

			An iterator over the children of this container.
		"""
		if include_self:
			yield self
		for child in self.children:
			yield child
			if recursive:
				yield from child.get_children(recursive=recursive)

	###########################################################################
	def get_child_by_name(self, name, recursive=True):
		""" Searches for a child container with the given name.

			# Arguments

			name: str. The name of the child container.
			recursive: bool (default: True). If True, each child container's
				children are also searched for the given named child.

			# Return value

			If the named child is found, it is returned. Otherwise, None is
			returned.
		"""
		for child in self.children:
			if child.name == name:
				return child
			elif recursive:
				result = child.get_child_by_name(name, recursive=recursive)
				if result is not None:
					return result
		return None

	###########################################################################
	def get_root(self):
		""" Returns the root node of the parent/child tree.
		"""
		if self.parent:
			return self.parent.get_root()
		else:
			return self

	###########################################################################
	def get_relative_by_name(self, name):
		""" Searches the entire parent/child hierarchy for a container with the
			given name.

			# Arguments

			name: str. The name of the container to search for.

			# Return value

			If the container is found, it is returned; otherwise, None is
			returned.
		"""
		return self.get_root().get_child_by_name(name, recursive=True)

	###########################################################################
	def parse(self, engine):
		""" Convenience function for parsing the container.

			This should not be overriden in derived classes. Override
			`_parse()` instead.
		"""
		if self._parsed:
			return

		self._parse_pre(engine)
		# TODO: Apply any "vars" that got parsed out during `_parse_pre()`.
		self._parse(engine)
		self._parse_post(engine)

		self._parsed = True

	###########################################################################
	def build(self, backend, rebuild=False):
		""" Convenience function for building the underlying operations.

			This should not be overriden in derived classes. Override
			`_build()` instead.
		"""
		if not self._parsed:
			raise ParsingError('Container must be parsed before being built.')
		if self._built is None or rebuild:
			self._built = list(self._build(backend))
		yield from self._built

	###########################################################################
	def _parse_pre(self, engine):
		""" Pre-parsing hook.

			Specifically, we may need to parse out the "vars" field (if
			present) so that we can do variable substitution during `_parse()`.
		"""
		if isinstance(self.data, str):
			self.data = engine.evaluate(self.data)
			if isinstance(self.data, str):
				self.data = {self.data : None}

	###########################################################################
	def _parse_post(self, engine):
		""" Post-parsing hook.

			The primary purpose of this is to register the container with the
			templating engine so that it can be referenced by other containers
			or specification sections.
		"""
		engine.register(self)

	###########################################################################
	def _parse(self, engine):
		""" Parse the container.

			This should be overriden in derived classes.
		"""
		if 'tag' in self.data:
			self.tags = engine.evaluate(self.data['tag'], recursive=True)
			if not isinstance(self.tags, (list, tuple)):
				self.tags = [self.tags]
		else:
			self.tags = []

		if self.name is None:
			if 'name' in self.data:
				self.name = engine.evaluate(self.data['name'])
			else:
				self.name = self.unique_name()

		if 'oldest' in self.data:
			self.oldest = engine.evaluate(self.data['oldest'], recursive=True)
			if not isinstance(self.oldest, (list, tuple)):
				self.oldest = [self.oldest]
		else:
			self.oldest = []

		container_name = self.get_container_name()
		if container_name in self.data:
			self.args = engine.evaluate(self.data[container_name])
		else:
			warnings.warn('This container has no arguments.', SyntaxError)
			self.args = None

		if 'inputs' in self.data:
			self.inputs = engine.evaluate(self.data['inputs'], recursive=True)
			if not isinstance(self.inputs, (list, tuple)):
				self.inputs = [self.inputs]
		else:
			self.inputs = []

		if 'sink' in self.data:
			sink = engine.evaluate(self.data['sink'])
			if isinstance(sink, str):
				if sink.lower() in ('yes', 'true', 'on'):
					sink = True
				elif sink.lower() in ('no', 'false', 'off'):
					sink = False
				else:
					raise ParsingError('Cannot evaluate boolean "sink" '
						'string: {}'.format(sink))
			elif not isinstance(sink, bool):
				raise ParsingError('Cannot evaluate boolean "sink": {}'.format(
					sink))

			self.sink = sink
		else:
			self.sink = False

	###########################################################################
	def _build(self, backend):
		""" Constructs the backend-specific data objects.

			# Arguments

			backend: Backend instance. The backend to use for constructing the
				container.

			# Return value

			An iterator that yields Layers.
		"""
		raise NotImplementedError

	###########################################################################
	def validate(self, key, required=True, dtype=None, list_like=False,
		raise_error=True):
		""" Checks if a key exists and optionally has the right type.

			# Arguments

			key: object. The key to check for in the dictionary.
			required: bool (default: True). Whether or not the absence of the
				key should be a fatal error.
			dtype: type or tuple of types (default: None). If not None and the
				key exists, then the value is checked to see if its type
				matches `dtype` or, if `dtype` is a tuple, the value is checked
				to see if its type matches one of the types in `dtype`.
			list_like: bool or type or list of types (default: False). Only
				used when `dtype` is not None. If True, then the value is
				assumed to be iterable, and type-checking takes place over each
				element of the iterable rather than to the iterable type
				itself. If `list_like` is a type or list of types, then the
				iterable is also checked to ensure that it is one of the
				supported types, and then `dtype` type-checking still occurs
				over each element of the iterable.
			raise_error: bool (default: True). If True, validation failures
				raise a ParsingError exception; otherwise, a boolean result is
				returned.

			# Return value

			If `raise_error` is False, then True is returned if validation
			passes, and False is returned otherwise.

			# Exceptions

			If the key is required but missing, or if the key is present by has
			an incorrect type, then a ParsingError is thrown if `raise_error`
			is True.
		"""
		name = self.__class__.__name__

		# First check if the key exists.
		if key not in self.data:
			# If the key is not required, then we're done.
			if not required:
				return True

			if raise_error:
				raise ParsingError('Missing key for {}: {}'.format(name, key))
			else:
				return False

		# Now do type checking if requested.
		if dtype is not None:
			if list_like is True:
				if all(isinstance(val, dtype) for val in self.data[key]):
					return True
			elif list_like:
				if isinstance(self.data[key], list_like):
					if all(isinstance(val, dtype) for val in self.data[key]):
						return True
			else:
				if isinstance(self.data[key], dtype):
					return True

			if raise_error:
				raise ParsingError('Bad type for key "{}" in "{}". Expected '
					'{}{} but received {}: {}'.format(key, name,
					'a list of ' if list_like else '',
					dtype, type(self.data[key]), self.data[key])
				)
			else:
				return False

		return True

	###########################################################################
	def terminal(self):
		""" Whether or not this container is responsible for producing backend-
			specific operations.

			# Return value

			If this container actually produces backend-specific operations (as
			opposed to deferring to children for specific layers), or if it
			wraps backend operations with additional operations, then this
			function should return True.

			If this container simply yields its children's containers and
			provides higher-level control of the architecture, then it should
			return False.
		"""
		raise NotImplementedError

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import warnings
from collections import namedtuple, OrderedDict, deque
from ..containers import Container
from ..containers.operators import ContainerGroup
from ..containers.layers import Placeholder
from ..engine import PassthroughEngine

logger = logging.getLogger(__name__)

# Convenience class for keeping track of high-level network nodes.
ContainerNode = namedtuple('ContainerNode',
	['inputs', 'container', 'sink', 'outputs']
)

###############################################################################
class Node:							# pylint: disable=too-few-public-methods
	""" Class for representing nodes in the low-level neural network graph.

		Each Node instance represents one primitive tensor operation.
	"""

	# Counter for tracking node IDs (used for debugging purposes).
	counter = 0

	###########################################################################
	def __init__(self, inputs=None, operation=None, value=None, outputs=None):
		""" Creates a new Node.

			# Arguments

			inputs: list of Nodes. The Nodes that are inputs to this operation.
			operation: object. The tensor operation to perform; the type is
				backend-specific.
			value: object. The output of the tensor operation, or None to
				indicate that it hasn't been evaluated yet.
			outputs: list of Nodes. The Nodes that use this node as input.
		"""
		self.inputs = inputs
		self.operation = operation
		self.value = value
		self.outputs = outputs

		# Increment and store the Node ID counter for debugging purposes.
		self.node_id = Node.counter
		Node.counter += 1

	###########################################################################
	def __str__(self):
		""" Return a string representation of this Node, suitable for
			debugging.
		"""
		return 'Node(id={node_id}, inputs={inputs}, operation={operation}, ' \
			'value={value}, outputs={outputs})'.format(
				node_id=self.node_id,
				inputs=[node.node_id for node in self.inputs]
					if self.inputs else None,
				operation=self.operation,
				value=self.value,
				outputs=[node.node_id for node in self.outputs]
					if self.outputs else None
			)

###############################################################################
class ExtensionCallback:
	""" Callback interface for Model's extension callback hooks.
	"""
	###########################################################################
	def extended(self, name):
		""" Callback after an extension is added to a model.

			# Arguments

			name: str. The name of the extension that was added.
		"""
		raise NotImplementedError

	###########################################################################
	def retracted(self, name):
		""" Callback after an extension is removed from a model.

			# Arguments

			name: str. The name of the extension that was removed.
		"""
		raise NotImplementedError

###############################################################################
class ExtensionState(ExtensionCallback):\
	# pylint: disable=too-few-public-methods
	""" Context management for saving the state of the model's extensions.

		This is useful for allowing an arbitrary number of temporary changes to
		the model that should be discarded when the context is exited.
	"""

	###########################################################################
	def __init__(self, model):
		""" Create a new ExtensionState.

			# Arguments

			model: Model instance. The model to track the extension state of.
		"""
		self.model = model
		self.changed = False
		self.extensions = OrderedDict()

	###########################################################################
	def __enter__(self):
		""" Enter context management and keep track of the extensions that
			change.
		"""
		logger.debug('Entering a model state context.')
		self.extensions = OrderedDict(self.model.extensions.items())
		self.changed = False
		self.model.add_extension_callback(self)
		return self

	###########################################################################
	def __exit__(self, exc_type, exc_value, traceback):
		""" Exit context management and restore the extension state.
		"""
		logger.debug('Leaving a model state context.')
		self.model.remove_extension_callback(self)

		if self.changed:
			logger.debug('Reverting changes during extension state '
				'monitoring.')

			# Remove all current extensions.
			for name in reversed(self.model.extensions):
				self.model.retract(name, rebuild=False)

			# Add all the previous extensions.
			for name, containers in self.extensions.items():
				self.model.extend(name, containers, rebuild=False)

			# Build the model.
			self.model.build()
		else:
			logger.debug('Nothing changed during extension state monitoring.')

	###########################################################################
	def extended(self, name):
		""" Callback after an extension is added to a model.
		"""
		self.changed = True

	###########################################################################
	def retracted(self, name):
		""" Callback after an extension is removed from a model.
		"""
		self.changed = True

###############################################################################
class Extension:					# pylint: disable=too-few-public-methods
	""" Context management for model extension.

		This is used to temporarily add an extension to a model. The extension
		is added when the context is entered, and removed when the context is
		exited.
	"""

	###########################################################################
	@staticmethod
	def _token(x):
		""" Produces random strings of hexadecimal digits.

			If the 'secrets' module is present (Python 3.6+), it is used.
			Otherwise, 'random' is used.

			# Arguments

			x: int. The number of bytes to generate.

			# Return value

			A string of length 2*n.
		"""
		try:
			import secrets
			return secrets.token_hex(x)
		except ImportError:
			import hashlib
			import random
			state = hashlib.sha256()
			result = ''
			while len(result) < x*2:
				state.update(bytes(random.randint(0, 255) for i in range(100)))
				result += state.hexdigest()
			return result[:x*2]

	###########################################################################
	def __init__(self, model, containers, name=None):
		""" Creates a new extension.

			# Arguments

			model: Model instance. The model to modify.
			containers: list of Containers. The containers to extend the model
				with.
			name: str (default: None). The name of the extension. If None, a
				random name is used.
		"""
		self.model = model
		self.containers = containers
		self.name = name or Extension._token(16)

	###########################################################################
	def __enter__(self):
		""" Enter the context manager and apply the extension.
		"""
		self.model.extend(self.name, self.containers)
		return self

	###########################################################################
	def __exit__(self, exc_type, exc_value, traceback):
		""" Exit the context manager and remove the extension.
		"""
		self.model.retract(self.name)

###############################################################################
class Model:
	""" Class for holding the model corresponding to a given container.

		Unlike a container, a model's layers are connected vis-a-vis the
		backend, and it has tracked input and output connections.
	"""

	###########################################################################
	def __init__(self, backend, containers):
		""" Create a new model.

			# Arguments

			backend: Backend instance. The backend to use in assembling this
				model.
			container: Container or list of Containers. The container(s) to
				assemble in a model. If this is a list, then note that order is
				important, since layers without explicit 'input' connections
				will be connected to the previous layer.
		"""
		if isinstance(containers, Container):
			containers = [containers]
		self.root = ContainerGroup(containers)

		self.backend = backend

		self.extensions = OrderedDict()
		self.extension_callbacks = []

		self.network = None
		self.inputs = None
		self.outputs = None
		self.endpoints = None
		self.container_nodes = None

		self._parsed = False

	###########################################################################
	def add_extension_callback(self, extension_callback):
		""" Adds an extension callback.

			Extension callbacks are called just after a new extension is added
			and immediately before an existing extension is removed.

			# Arguments

			extension_callback: ExtensionCallback instance. The callback to
				register.
		"""
		self.extension_callbacks.append(extension_callback)

	###########################################################################
	def remove_extension_callback(self, extension_callback):
		""" Removes an extension callback.

			# Arguments

			extension_callback: ExtensionCallback instance. The callback to
				register.
		"""
		self.extension_callbacks.remove(extension_callback)

	###########################################################################
	def save(self, filename):
		""" Saves the model weights to the given filename.

			# Arguments

			filename: str. The filename to write the weights to.

			# Notes

			The file format is backend-specific. There is no guarantee of
			compatability between different backends.
		"""
		logger.debug('Saving model weights to: %s', filename)
		self.backend.save(self, filename)

	###########################################################################
	def restore(self, filename):
		""" Load the model weights from the given filename.

			# Arguments

			filename: str. The filename to read the weights from.

			# Notes

			The file format is backend-specific. There is no guarantee of
			compatability between different backends.
		"""
		logger.debug('Loading model weights from: %s', filename)
		self.backend.restore(self, filename)

	###########################################################################
	def apply_provider_knowledge(self, provider):
		""" Enables inference of tensor shapes by modifying the parsed
			containers given provider information.
		"""
		logger.debug('Applying provider-inferred shapes to input layers.')
		if provider.keys is None:
			logger.debug('No provider keys available. Cannot infer shapes.')
			return

		sources = dict(zip(provider.keys, provider.sources))
		for target in self.root.get_children(
				recursive=True, include_self=True
			):
			if not isinstance(target, Placeholder):
				continue

			logger.debug('Trying to infer shape for input "%s"', target.name)

			if target.name not in sources:
				logger.warning('Could not find a data source for model '
					'input "%s". Maybe you meant one of these: %s',
					target.name, ', '.join(provider.keys()))
				continue

			source = sources.pop(target.name)
			shape = source.shape()

			if target.shape is None:
				logger.debug('Inferred shape for input "%s": %s',
					target.name, shape)
				target.set_shape(shape)

			elif target.shape != shape:
				logger.warning('The input placeholder "%s" in the model '
					'has an explicit shape %s which disagrees with the '
					'shape of the corresponding data source %s.',
					target.name, target.shape, source.shape())

			else:
				logger.debug('Input "%s" already has a shape.', target.name)

	###########################################################################
	def parse(self, engine):
		""" Parses the model.
		"""
		if not self._parsed:
			if engine is None:
				logger.info('Creating a dummy engine for parsing the model.')
				engine = PassthroughEngine()
			self.root.parse(engine)
			self._parsed = True

	###########################################################################
	@staticmethod
	def _assemble(container, inputs, previous, outputs, sink, result):
		""" Enumerates all pieces of the model that must be built.

			This basically enumerates all containers, assigns them their
			resolved names, and tracks their inputs.

			# Arguments

			container: Container instance. The container to assemble
				ContainerNodes for.
			inputs: list or None. If a list, it should be a list of strings
				that indicate which containers are used as input to
				`container`.
			previous: str or None. If not None, then it is the name of the
				most recent container that was assembled, and which would serve
				as input to this container in the absence of any other explicit
				inputs.
			sink: bool. Whether or not the container is a sink.
			outputs: list or None. If a list, it should be a list of strings
				that indicate which names are associated with this
				`container`'s output. There may be multiple names due to nested
				containers.
			result: list. The output list to populate with ContainerNodes.

			# Return value

			The name of the most recently assembled container, suitable for use
			as `previous` in a recurrent/iterative call to `_assemble`.
		"""

		# If the order is `inputs or container.inputs` then the outermost
		# input list is used. If the order is `container.inputs or inputs`
		# then the innermost input list is used.
		inputs = inputs or container.inputs or previous
		if inputs is None:
			inputs = []
		elif not isinstance(inputs, (list, tuple)):
			inputs = [inputs]

		outputs = outputs or []
		if container.name:
			outputs = outputs + [container.name]

		logger.debug('Assembling container: %s', container.name)
		logger.debug('  Inputs: %s', inputs)
		logger.debug('  Outputs: %s', outputs)
		logger.debug('  Previous: %s', previous)

		if container.terminal():
			result.append(ContainerNode(
				inputs=tuple(inputs),			# List of names.
				container=container,			# Container instance.
				sink=sink or container.sink,	# If the container is a sink
				outputs=tuple(outputs)			# List of names.
			))
			return container.name

		else:
			children = list(container.get_children(recursive=False))
			for child in children:
				# Only the very last container should get tagged with the
				# parent container's output name.
				first_child = child == children[0]
				last_child = child == children[-1]
				previous = Model._assemble(
					container=child,
					inputs=inputs if first_child else previous,
					previous=previous,
					outputs=outputs if last_child else [],
					sink=(sink or container.sink) if last_child else False,
					result=result
				)
			return previous

	###########################################################################
	def assemble(self):
		""" Convenience function for constructing all the container-level graph
			nodes.

			# Return value

			A list of ContainerNodes.
		"""
		result = []
		Model._assemble(
			container=self.root,
			inputs=None,
			previous=None,
			outputs=None,
			sink=False,
			result=result
		)
		return result

	###########################################################################
	@staticmethod
	def _attach_within_container(backend, container_nodes):
		""" Instantiates the subgraphs of the low-level network that are
			trivially described within a container.

			# Arguments

			backend: Backend instance. The backend to use for instantiating
				graph elements.
			container_nodes: list of ContainerNodes. The list of assembled
				containers (as returned by `assemble()`) to create connections
				for.

			# Return value

			A tuple `(network, endpoints)`, where `network` is a list of
			partially constructed Nodes describing the network, and `endpoints`
			is a dictionary containing unlinked Nodes. `endpoints` has two
			keys: 'input' and 'output', which indicate whether the nodes
			contained in the respective keys are associated with the input or
			output ends of the container. The values in `endpoint` are also
			dictionaries of Nodes indexed by container name.
		"""

		network = []
		endpoints = {
			'input' : {},
			'output' : {}
		}

		# For each container, assemble its nodes.
		for container_node in container_nodes:
			logger.debug('Processing container node: %s', container_node)

			prev = None
			for layer in container_node.container.build(backend):
				node = Node(
					inputs=[prev] if prev is not None else None,
					operation=layer,
					value=None,
					outputs=[]
				)
				if prev is None:
					endpoints['input'][container_node.container.name] = node
					logger.debug('  Input: %s', node)
				else:
					prev.outputs.append(node)
					logger.debug('  Middle: %s', node)
				prev = node
				network.append(node)
			if prev is not None:
				endpoints['output'][container_node.container.name] = node
				logger.debug('  Output: %s', node)

		return network, endpoints

	###########################################################################
	@staticmethod
	def _attach_between_containers(container_nodes, endpoints):
		""" Finishes initializing low-level graph Nodes by enumerating
			connections between (rather than within) containers.

			# Arguments

			container_nodes: list of ContainerNodes. The ContainerNodes to
				create connections between.
			endpoints: dict. A dictionary describing the partially constructed
				Nodes, as returned by `_attach_within_container`.

			# Return value

			A tuple `(inputs, outputs)` where each of `inputs` and `outputs`
			are `OrderedDict' instances. These ordered dictionaries have keys
			which are the names of the containers associated with the inputs
			and outputs of the network, respectively. The values of these
			dictionaries are the Node instances themselves.
		"""
		# Data structures for holding the return values.
		inputs = OrderedDict()
		outputs = OrderedDict()

		# Find nodes which produce a given output name.
		node_by_output = OrderedDict(
			[(name, node) for node in container_nodes for name in node.outputs]
		)

		# Keep track of which nodes we use as inputs, so that later on we can
		# decide if they are output nodes.
		input_used = set()

		# Loop over each container node.
		for container_node in container_nodes:

			# Grab its name.
			container_name = container_node.container.name

			# Check if we have inputs to attach.
			if container_name not in endpoints['input']:
				warnings.warn('Container name was not found in list of open '
					'inputs. This is a bug, but we can try to skip it.')
				continue

			# Grab the node object.
			node = endpoints['input'][container_name]

			# Figure out where the input comes from.
			if isinstance(container_node.container, Placeholder):
				# Oh, it's just a placeholder. Sweet.
				inputs[container_name] = node
			else:
				# It's another container. Let's connect to each of its inputs.
				prev = []
				for input_name in container_node.inputs:

					# Was this created by another layer?
					if input_name in node_by_output:
						# Yes, attach to it.

						# Get the node
						input_node = node_by_output[input_name]

						# Mark its output as used by someone.
						input_used.add(input_node.container.name)

						# Now grab the node associated with the
						# input container's output.
						input_node = \
							endpoints['output'][input_node.container.name]

						# Add it as an input
						prev.append(input_node)

						# Let the input node know that we are one of its
						#outputs.
						input_node.outputs.append(node)

					else:
						# No other container owns this input. It's probably
						# because it is an implicit input, but without a
						# Placeholder to mark it as such.
						raise ValueError('Placeholder inference is not '
							'supported yet.')

				# Set the node's inputs.
				node.inputs = prev

		# Now figure out which nodes are truly outputs.
		for container_name in endpoints['output']:
			# Grab the node
			node = endpoints['output'][container_name]

			# Grab the container.
			container_node = node_by_output[container_name]

			# To be an output, the container needs to be a sink, or the output
			# has to not be used as an input anywhere.
			if container_node.sink or not container_name in input_used:
				outputs[container_name] = node

		return inputs, outputs

	###########################################################################
	def is_built(self):
		""" Returns True if this model has been built at some point.
		"""
		return self.network is not None

	###########################################################################
	def build(self):
		""" Builds the model.
		"""

		logger.info('Building the model.')

		if not self._parsed:
			logger.warning('The model has not been parsed yet. We will try to '
				'parse it without context, but this may easily fail. Make '
				'sure Model.parse() is called before Model.build().')
			self.parse(None)

		# Construct the high-level network nodes.
		container_nodes = self.assemble()

		# Assemble the in-container nodes.
		network, endpoints = Model._attach_within_container(
			self.backend,
			container_nodes
		)

		if logger.isEnabledFor(logging.DEBUG):

			logger.debug('Network:')
			for entry in network:
				logger.debug('  - %s', entry)

		# Now do the between-container attaching.
		inputs, outputs = Model._attach_between_containers(
			container_nodes,
			endpoints
		)

		logger.info('Model inputs:  %s', ', '.join(node for node in inputs))
		logger.info('Model outputs: %s', ', '.join(node for node in outputs))

		self._connect_network(self.backend, inputs)

		self.network = network
		self.inputs = inputs
		self.outputs = outputs
		self.endpoints = endpoints
		self.container_nodes = container_nodes

	###########################################################################
	def has_extension(self, name):
		""" Returns True if the model has the current extension attached.

			# Arguments

			name: str. The name of the extension to query.
		"""
		return name in self.extensions

	###########################################################################
	def extend(self, name, containers, rebuild=True):
		""" Extends the model with additional nodes.

			# Arguments

			name: str. The name of this extension, which can be used to remove
				it later. If an extension already exists with this name, a
				ValueError exception is raised.
			containers: list of Containers. The containers to extend the model
				with.
			rebuild: bool (default: True). Whether or not to rebuild the model
				after the extension is added. Generally, this is a very good
				idea, and should only be False if `build()` will be called
				explicitly (as when modifying extensions in a batch).
		"""
		if name in self.extensions:
			raise ValueError('An extension with the same name has already '
				'been applied: {}'.format(name))

		self.extensions[name] = containers
		for container in containers:
			self.root.add_child(container)

		if rebuild:
			self.build()

		for extension_callback in self.extension_callbacks:
			extension_callback.extended(name)

	###########################################################################
	def retract(self, name, rebuild=True):
		""" Removes a previously attached model extension.

			# Arguments

			name: str. The name of the extension to remove. If it doesn't
				exist, a ValueError is raised.
			rebuild: bool (default: True). Whether or not to rebuild the model
				after the extension is removed. Generally, this is a very good
				idea, and should only be False if `build()` will be called
				explicitly (as when modifying extensions in a batch).

			# Return value

			The list of Containers that previously extended the model.
		"""
		if name not in self.extensions:
			raise ValueError('No such extension found: {}'.format(name))

		containers = self.extensions.pop(name)
		for container in containers:
			self.root.remove_child(container)

		if rebuild:
			self.build()

		for extension_callback in self.extension_callbacks:
			extension_callback.retracted(name)

		return containers

	###########################################################################
	@staticmethod
	def _connect_network(backend, inputs):
		""" Applies all tensor operations to instantiate a fully-connected
			tensor graph.

			# Arguments

			backend: Backend instance. The backend to use to connect graph
				elements.
			network: list of Nodes. The nodes which constitute the network.
			inputs: list of Nodes. The input nodes of the network.
		"""
		queue = deque(inputs.values())
		while queue:
			node = queue.popleft()

			logger.debug('Connecting node: %s', node)

			if node.inputs is None:
				node.value = node.operation
			else:
				node.value = backend.connect(
					inputs=[x.value for x in node.inputs],
					target=node.operation
				)

			if node.outputs:
				queue.extend(node.outputs)

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy
from . import Supplier
from ..sources import VanillaSource

###############################################################################
class NumpyDictSupplier(Supplier):
	""" A supplier which parses out data from a Numpy file (*.npy) containing a
		dictionary whose values are the source names, and whose values are the
		Numpy arrays that should be returned.
	"""

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the supplier.
		"""
		return 'numpy_dict'

	###########################################################################
	def __init__(self, source, *args, **kwargs):
		""" Creates a new Numpy dictionary supplier.

			# Arguments

			source: str. Filename of Numpy file (*.npy) to load.
		"""
		super().__init__(*args, **kwargs)

		self.source = source
		self.data = None

	###########################################################################
	def _load(self):
		""" Loads the data (only if it hasn't already been loaded).
		"""
		if self.data is not None:
			return
		data = numpy.load(self.source)
		data = data[()]
		self.data = {k : VanillaSource(v) for k, v in data.items()}

	###########################################################################
	def get_sources(self, sources=None):
		""" Returns all sources from this provider.
		"""
		self._load()

		if sources is None:
			sources = list(self.data.keys())
		elif not isinstance(sources, (list, tuple)):
			sources = [sources]

		for source in sources:
			if source not in self.data:
				raise KeyError(
					'Invalid data key: {}. Valid keys are: {}'.format(
						source, ', '.join(str(k) for k in self.data.keys())
				))

		return {k : self.data[k] for k in sources}

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy

from ..utils import download_file, idx
from . import Supplier
from ..sources import VanillaSource

###############################################################################
class MnistSupplier(Supplier):
	""" A supplier which supplies MNIST image/label pairs. These are downloaded
		from the internet, verified, and parsed as IDX files.
	"""

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the supplier.
		"""
		return 'mnist'

	###########################################################################
	def __init__(self, labels, images, *args, **kwargs):
		""" Creates a new MNIST supplier.

			# Arguments

			labels: str or dict. If str, the path to the IDX file containing
				the image labels. If a dict, it should follow one of these
				formats:

				1. {"url" : URL, "sha256" : SHA256, "path" : PATH}, where URL
				   is the source URL (if the image file needs downloading),
				   SHA256 is the SHA-256 hash of the file (optional, and can be
				   missing or None to skip the verification), and PATH is the
				   directory to store the file in (if missing or None it
				   defaults to the system temporary directory).
				2. {"local" : FILE}, where FILE is the path to the IDX file.

			images: str or dict. Specifies where the MNIST images can be found.
				Accepts the same values as `labels`.
		"""
		super().__init__(*args, **kwargs)

		self.data = {
			'images' : MnistSupplier._normalize(
				VanillaSource(idx.load(MnistSupplier._get_filename(images)))
			),
			'labels' : MnistSupplier._onehot(
				VanillaSource(idx.load(MnistSupplier._get_filename(labels)))
			)
		}

	###########################################################################
	@staticmethod
	def _onehot(source):
		onehot = numpy.zeros((len(source), 10))
		for i, row in enumerate(source.data):
			onehot[i][row] = 1
		source.data = onehot

		return source

	###########################################################################
	@staticmethod
	def _normalize(source):
		# Numpy won't automatically promote the uint8 fields to float32.
		source.data = source.data.astype(numpy.float32)

		# Normalize
		source.data /= 255
		source.data -= source.data.mean()

		source.data = numpy.expand_dims(source.data, axis=-1)

		return source

	###########################################################################
	@staticmethod
	def _get_filename(target):
		""" Returns the filename associated with a particular target.

			# Arguments

			target: str or dict. The target specification. For locally-stored
				files, it can be a string (path to file) or a dictionary with
				key 'local' that contains the file path. For network files,
				it is a dictionary with 'url' (source URL); it may also
				optionally contain 'sha256' (SHA256 checksum) and 'path' (local
				storage directory for the file).

			# Return value

			String to the file's locally stored path. May not exist.
		"""

		if isinstance(target, str):
			return target
		elif isinstance(target, dict):
			if 'url' in target:
				return download_file(
					url=target['url'],
					sha256=target.get('sha256'),
					target_dir=target.get('path')
				)
			elif 'local' in target:
				return target['local']
			else:
				raise ValueError('Expected either "url" (for downloading data '
					'sources) or "local" (for locally-stored sources), but '
					'neither key was found in the MNIST specification: {}'
					.format(target))
		else:
			raise ValueError('Unexpected data type for MNIST target: {}'
				.format(target))

	###########################################################################
	def get_sources(self, sources=None):
		""" Returns all sources from this provider.
		"""

		if sources is None:
			sources = list(self.data.keys())
		elif not isinstance(sources, (list, tuple)):
			sources = [sources]

		for source in sources:
			if source not in self.data:
				raise KeyError(
					'Invalid data key: {}. Valid keys are: {}'.format(
						source, ', '.join(str(k) for k in self.data.keys())
				))

		return {k : self.data[k] for k in sources}

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
from kur.utils import get_subclasses

###############################################################################
class Reader:
	""" Base class for all readers.

		Readers are responsible for reading files and constructing parsable
		Python objects out of the encoded data. For an example, you might have
		a JSON reader, a YAML reader, etc.
	"""

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the reader.

			# Return value

			A lower-case string unique to this reader.
		"""
		return cls.__name__.lower()

	###########################################################################
	@staticmethod
	def get_all_readers():
		""" Returns all Reader subclasses.
		"""
		for cls in get_subclasses(Reader):
			yield cls

	###########################################################################
	@staticmethod
	def get_reader_by_name(name):
		""" Finds a reader class with the given name.
		"""
		name = name.lower()
		for cls in Reader.get_all_readers():
			if cls.get_name() == name:
				return cls
		raise ValueError('No such reader with name "{}"'.format(name))

	###########################################################################
	@classmethod
	def supported_filetypes(cls):
		""" Returns a list of supported file extensions.

			# Return value

			A tuple of lowercase extensions (without the period) that generally
			indicate that the Reader subclass can handle the implied file type.
			If no file extensions is particularly indicative, then an empty
			tuple can be returned. If the absence of a file extension is
			particularly indicative, then an empty string can be present in the
			tuple.
		"""
		raise NotImplementedError

	###########################################################################
	@staticmethod
	def get_reader_for_file(filename):
		""" Returns the first Reader that claims to be able to read the given
			file.

			# Arguments

			filename: str. The filename to find a Reader for.

			# Return value

			A class that can read the given filename. If no such class can be
			found, a ValueError is raised.
		"""
		_, ext = os.path.splitext(filename)
		ext = ext.lower()
		if ext.startswith('.'):
			ext = ext[1:]

		for cls in Reader.get_all_readers():
			if ext in cls.supported_filetypes():
				return cls

		raise ValueError(
			'No such reader could be found for file: {}'.format(filename))

	###########################################################################
	@staticmethod
	def read_file(filename):
		""" Convenience function for reading data from a file.
		"""
		reader = Reader.get_reader_for_file(filename)()
		with open(filename) as fh:
			return reader.read(fh.read())

	###########################################################################
	def read(self, data):
		""" Reads the data and returns a native Python dictionary.

			# Arguments

			data: str. The data string to parse.

			# Return value

			A Python dictionary or list representing the data.
		"""
		raise NotImplementedError

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
from ..utils import get_subclasses

logger = logging.getLogger(__name__)

###############################################################################
class Provider:						# pylint: disable=too-few-public-methods
	""" Base class for all data providers.

		Data providers take a set of data sources and synchronize their inputs,
		providing batches to the network for training. Introducing the Provider
		as an abstraction between a data source and the training loop is
		convenient in order to allow different data sources to produce data at
		different rates (e.g., in-memory, on-disk, over-network).

		From a software design perspective, the Provider is a strategy pattern
		for allowing the training system to retrieve data.

		# Expected behavior

		- Whenever any Source iterator finishes, the entire Provider stops
		  iterating.
		- Whenever a Provider stops iterating, that is the end of the epoch.
		- Whenever a Provider starts iterating (at the beginning of each
		  epoch), it creates new iterators for each Source. Source iterators
		  are not persisted across Provider iterations.
		- Sources are permitted to be different lengths.
		- Because the Provider stops iterating (and terminates the epoch) as
		  soon as the first (shortest) Source completes, and because Sources
		  are allowed to have different lengths, some sources (which are finite
		  but longer than the shortest source) may never have their entire
		  datasets iterated over (because the epoch will end before the last
		  entries in the longer sources are reached). If this is unacceptable,
		  then you should consider restructuring your data sources, or using
		  Shuffleable sources with a ShuffleProvider.
		- Sources are said to be infinite if they cannot determine how many
		  samples they can provide (i.e., "len" returns None for the Source).
		  A Provider is said to be infinite if all of its Sources are infinite.
		  Note that an infinite Provider or Source does not need to never
		  terminate; it simply means that it cannot determine when the iterator
		  will finish. An infinite Source can still stop iterating, causing the
		  infinite Provider to complete data providing for the epoch.
	"""

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the provider.

			# Return value

			A lower-case string unique to this provider.
		"""
		return cls.__name__.lower()

	###########################################################################
	@staticmethod
	def get_all_providers():
		""" Returns an iterator to the names of all providers.
		"""
		for cls in get_subclasses(Provider):
			yield cls

	###########################################################################
	@staticmethod
	def get_provider_by_name(name):
		""" Finds a provider class with the given name.
		"""
		name = name.lower()
		for cls in Provider.get_all_providers():
			if cls.get_name() == name:
				return cls
		raise ValueError('No such provider with name "{}"'.format(name))

	###########################################################################
	def __init__(self, sources):
		""" Create a new data provider.

			# Arguments

			sources: dict or list. If this is a list, then it is a list of
				Sources. If this is a dictionary, its values are Sources and
				its keys are string names that are used by the training process
				to determine which nodes in the network should receive which
				information.
		"""
		if isinstance(sources, list):
			self.keys = None
			self.sources = sources
		elif isinstance(sources, dict):
			self.keys = list(sources.keys())
			self.sources = [sources[k] for k in self.keys]
		else:
			raise ValueError('Unknown or unexpected type for "sources": {}'
				.format(type(sources)))

		lens = [len(source) for source in self.sources]
		len_set = {i for i in lens if i is not None}
		if len(len_set) == 1:
			# All finite sources are the same length.
			self.entries = len_set.pop()
		elif len_set:
			# Finite sources have differing lengths.
			self.entries = min(lens)
			if logger.isEnabledFor(logging.INFO):
				if self.keys is None:
					msg = 'Sizes are: {}'.format(
						', '.join(str(x) for x in lens)
					)
				else:
					msg = 'Sizes are: {}'.format(
						', '.join('{}={}'.format(k, v)
							for k, v in zip(self.keys, lens)
						)
					)
				logger.info('Mixed source lengths. Make sure you know what '
					'you are doing. %s', msg)
		else:
			# There are no finite sources.
			self.entries = None
		self.lens = lens

	###########################################################################
	def pre_iter(self):
		""" This is a hook that is called whenever the provider begins
			generating data for the next epoch.
		"""
		pass

	###########################################################################
	def __iter__(self):
		""" Return an iterator which can generate batches of data.

			The iterator should return a tensor of shape `(X, ) +
			self.shape()`, where `X` is number of entries in the batch returned
			by this provider (it is implementation-specific, but will be used
			by the training/testing/evauation process as a batch).

			# Notes:

			- The implementation must call `pre_iter()` as the first call in
			  `__iter__`.
		"""
		raise NotImplementedError

	###########################################################################
	def wrap(self, data):
		""" Wraps data up in the same way it was presented to the Provider.

			# Arguments

			data: list. A list of tensors, one for each data source.

			# Return value

			If `sources` in `__init__` was a list, returns a list of tensors
			in the same order. If `sources` was a dict, returns a dictionary
			with the same keys.
		"""
		if self.keys:
			return {k : v for k, v in zip(self.keys, data)}
		else:
			return data

	###########################################################################
	def __len__(self):
		""" Returns the total number of entries that this provider returns per
			epoch.

			# Return value

			If the provider is finite (because at least one of its sources is
			finite), then the provider must return the number of samples it
			provides per epoch (that is, the length of the shortest finite
			source).

			If the provider is infinite (because all of its sources are
			infinite), then it should return None.
		"""
		return self.entries

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging

# Get the eight standard colors.
BLACK, RED, GREEN, YELLOW, BLUE, MAGENTA, CYAN, WHITE = range(8)

# ANSI escape sequences for colored outputs.
RESET_SEQ = "\033[0m"
COLOR_SEQ = "\033[1;{}m"
BOLD_SEQ = "\033[1m"

# Color codes for each log-level.
COLORS = {
	'DEBUG': BLUE,
	'INFO': WHITE,
	'WARNING': YELLOW,
	'ERROR': RED,
	'CRITICAL': RED
}

###############################################################################
def basicConfig(**kwargs):						# pylint: disable=invalid-name
	""" A convenience function which mimics the behavior of
		`logging.basicConfig()`, but creates a color formatter.
	"""
	logger = logging.getLogger()
	if logger.hasHandlers():
		return

	if sum(k in kwargs for k in ('filename', 'handlers', 'stream')) > 1:
		raise ValueError('Only one of "filename", "stream", and "handlers" '
			'can be specified.')

	# Create the formatter
	formatterArgs = {							# pylint: disable=invalid-name
		k : kwargs[k] for k in ('datefmt', 'style') if k in kwargs
	}
	if 'format' in kwargs:
		formatterArgs['fmt'] = kwargs['format']
	formatter = ColorFormatter(**formatterArgs)

	# Create a new handler that writes to stderr, and attach the
	# color formatter to it.
	if 'filename' in kwargs:
		handlers = [logging.FileHandler(
			kwargs['filename'],
			mode=kwargs.pop('filemode', 'a')
		)]
	elif 'handlers' in kwargs:
		handlers = kwargs['handlers']
	else:
		handlers = [logging.StreamHandler(stream=kwargs.pop('stream', None))]

	for handler in handlers:
		handler.setFormatter(formatter)
		logger.addHandler(handler)

	# Set the log-level
	if 'level' in kwargs:
		logger.setLevel(kwargs['level'])

###############################################################################
class ColorFormatter(logging.Formatter):
	""" A formatter for Python's logging framework which can produce colored
		logging statements.

		# Usage

		Normally, a basic logger can be configured with:

		```python
		import logging

		logging.basicConfig(
			level=(logging.DEBUG if args.verbose else logging.INFO),
			format='[{asctime} {name}:{lineno}] {message}',
			style='{'
		)
		```

		For colored output, the custom formatter must be created. This
		requires a more complex setup:

		```python
		import logging
		from logcolor import ColoredFormatter

		# Get the root logger.
		logger = logging.getLogger()

		# Create a new handler that writes to stderr, and attach the
		# color formatter to it.
		stream = logging.StreamHandler()
		stream.setFormatter(ColorFormatter(
			# $COLOR triggers the start of a colored sequence
			# $REST clears the coloring
			fmt='$COLOR[{levelname} {asctime} '
				'{name}:{lineno}]$RESET {message}',
			style='{'
		))
		logger.addHandler(stream)
		logger.setLevel(logging.DEBUG if args.verbose else logging.INFO)
		```

		A convenience function is supplied to make this easy:
		```python
		import logcolor

		logcolor.basicConfig(
			level=(logging.DEBUG if args.verbose else logging.INFO),
			format='$COLOR[{asctime} {name}:{lineno}]$RESET {message}',
			style='{'
		)
		```
	"""

	###########################################################################
	def format(self, record):
		""" Formats the log record using colors.
		"""
		# Get the message using the default formatter
		message = super().format(record)
		# Get the color corresponding to the log-level.
		levelname = record.levelname
		color = COLOR_SEQ.format(30 + COLORS[levelname])
		# Do the replacements to trigger colored outputs
		message = message.replace("$RESET", RESET_SEQ) \
			.replace("$BOLD", BOLD_SEQ).replace("$COLOR", color)
		# Return the result, and terminate it in a reset just to be safe.
		return message + RESET_SEQ

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from ..utils import get_subclasses

###############################################################################
class Loss:
	""" Base class for all loss functions (also called objective functions).
	"""

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the loss function.

			# Return value

			A lower-case string unique to this loss function.
		"""
		return cls.__name__.lower()

	###########################################################################
	@staticmethod
	def get_all_losses():
		""" Returns an iterator to the names of all loss functions.
		"""
		for cls in get_subclasses(Loss):
			yield cls

	###########################################################################
	@staticmethod
	def get_loss_by_name(name):
		""" Finds a loss function class with the given name.
		"""
		name = name.lower()
		for cls in Loss.get_all_losses():
			if cls.get_name() == name:
				return cls
		raise ValueError('No such loss function with name "{}"'.format(name))

	###########################################################################
	def __init__(self, weight=None):
		""" Creates a new loss function.

			# Arguments

			weight: float. A relative weight to apply to this loss function.
				This is only meaningful in models which have multiple loss
				functions.
		"""
		self.weight = 1.0 if weight is None else weight

	###########################################################################
	def get_weight(self):
		""" Returns the loss function weight.
		"""
		return self.weight

	###########################################################################
	def get_loss(self, backend):
		""" Returns the loss function that can be used by the implementation-
			specific model.
		"""
		raise NotImplementedError

	###########################################################################
	def modify(self, model, name): \
		# pylint: disable=unused-argument,no-self-use
		""" Modify/extend the model to fit this loss function.

			Some loss functions will want to modify the model in some way in
			order to properly instrument the model. For example, CTC loss is a
			little different in the sense that it wants additional inputs at
			training and evaluation time. Many loss functions, however, will
			not need this functionality.

			# Arguments

			model: Model instance. The model to modify.
			name: str. The output of the model to modify.

			# Return value

			The name of the model output to apply this loss function to.
			Normally, this is the same as `name`, but if the loss function
			modifies the model, then it may need to act on a different layer
			instead.
		"""
		return name

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from . import Operator
from . import ParsingError

###############################################################################
class ForLoop(Operator):				# pylint: disable=too-few-public-methods
	""" A classic "for" loop for creating more complex or arbitrary length
		models.

		# Example

		```
		for:
		  range: 10
		  with_index: i
		  iterate:
		    - debug: "Hello {{ i }}"
		```
	"""

	###########################################################################
	@classmethod
	def get_container_name(cls):
		""" Returns the name of the container class.

			Obviously, "for" is an overloaded word in programming. So rather
			than risk problems with statements like `import for`, it is better
			to just have a different name that is used in parsing containers.
		"""
		return 'for'

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Create a new for loop.
		"""
		super().__init__(*args, **kwargs)
		self.limit = None
		self.index = None

	###########################################################################
	def _parse(self, engine):
		""" Parse and construct the child containers.
		"""

		# Always call the parent.
		super()._parse(engine)

		# Parse self
		if 'range' not in self.args:
			raise ParsingError('Missing "range" key in for loop.')
		self.limit = engine.evaluate(self.args['range'])
		try:
			self.limit = int(self.limit)
		except ValueError:
			raise ParsingError('"limit" in a "for" loop must evaluate to an '
				'integer; got this instead: {}'.format(self.limit))

		if 'with_index' not in self.args:
			self.index = 'index'
		else:
			self.index = engine.evaluate(self.args['with_index'])

		if 'iterate' not in self.args:
			raise ParsingError('Missing "iterate" key in for loop.')
		target = engine.evaluate(self.args['iterate'])

		# Parse children
		for index in range(self.limit):
			with engine.scope(**{self.index : index}):
				for entry in target:
					self.new_child_from_data(entry).parse(engine)

	###########################################################################
	def _build(self, backend):
		""" Construct each child.
		"""
		for child in self.children:
			yield from child.build(backend)

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import warnings
from . import Layer, ParsingError

logger = logging.getLogger(__name__)

###############################################################################
class Placeholder(Layer):				# pylint: disable=too-few-public-methods
	""" A shape placeholder which serves as a size hint, typically used as the
		first layer for each input to the model.

		Placeholders are also convenient places to name an input stream.
	"""

	###########################################################################
	@classmethod
	def get_container_name(cls):
		""" Returns the name of the container class.
		"""
		return 'input'

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Creates a new placeholder.
		"""
		super().__init__(*args, **kwargs)
		self.shape = None

	###########################################################################
	def _parse_pre(self, engine):
		""" Pre-parsing hook.
		"""
		super()._parse_pre(engine)

		container_name = self.get_container_name()
		if container_name in self.data:
			data = engine.evaluate(self.data[container_name])
			if isinstance(data, str):
				if 'name' in self.data:
					name = engine.evaluate(self.data['name'])
					logger.warning('Conflicting naming schemes for '
						'placeholder: "%s" and "%s". Using: "%s".',
						data, name, name)
				else:
					logger.debug('Using short-hand name for placeholder: %s',
						data)
					self.name = data
		else:
			warnings.warn('Parsing oddity in placeholder: {}'.format(
				self.data
			))

	###########################################################################
	def set_shape(self, shape):
		""" Sets a shape.
		"""
		if self.shape is not None:
			logger.warning('Modifying the shape of Placeholder "%s".',
				self.name)

		if not isinstance(shape, (list, tuple)):
			shape = (shape, )
		for x in shape:
			if not isinstance(x, int):
				raise ParsingError(
					'All entries in "shape" must be integers. Shape is: {}'
					.format(shape)
				)
		self.shape = shape

	###########################################################################
	def _parse(self, engine):
		""" Parse the placeholder.
		"""

		super()._parse(engine)

		if 'shape' not in self.args:
			logger.debug('Placeholder "%s" has a deferred shape.', self.name)
		else:
			self.set_shape(engine.evaluate(self.args['shape'], recursive=True))

	###########################################################################
	def _build(self, backend):
		""" Create the backend-specific placeholder.
		"""
		if backend.get_name() == 'keras':

			if self.shape is None:
				raise ParsingError(
					'Placeholder "{}" requires a shape.'.format(self.name))

			import keras.layers as L			# pylint: disable=import-error
			yield L.Input(
				shape=self.shape,
				name=self.name
			)

		else:
			raise ValueError('Unknown or unsupported backend: {}'.format(backend))

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import signal
import logging

logger = logging.getLogger(__name__)

###############################################################################
class CriticalSection:				# pylint: disable=too-few-public-methods
	""" Protects critical code from system signals (e.g., keyboard interrupts).

		# Usage

		Prevent KeyboardInterrupt until after the critical section:
		```python
		with CriticalSection():
			# Some code ...

			# This interrupt will be suppressed.
			raise KeyboardInterrupt

			# This code will continue to execute
			# ...

		# Once out of the critical section, the suppressed interrupt will be
		# re-raised.
		```
	"""

	###########################################################################
	def __init__(self, signals=None):
		if signals is None:
			signals = signal.SIGINT

		if not isinstance(signals, (list, tuple)):
			signals = [signals]		# pylint: disable=redefined-variable-type

		self.signals = signals
		self.old_handlers = {}
		self.received = {}

	###########################################################################
	def _handle(self, sig, *args):
		if sig in self.received:
			self.received[sig].append(args)
			logger.debug('Signal received in critical section: %s', sig)
		else:
			logger.warning('Unexpected signal received: %s', sig)

	###########################################################################
	def __enter__(self):
		""" Enters a critical section.
		"""

		self.old_handlers = {}
		for sig in self.signals:

			def handler(sig=sig, *args):
				""" Signal handler (to avoid cell variable problems)
				"""
				self._handle(sig, *args)

			self.received[sig] = []
			self.old_handlers[sig] = signal.signal(
				sig,
				handler
			)

	###########################################################################
	def __exit__(self, exc_type, exc_value, traceback):
		""" Leaves a critical section.
		"""

		# Restore the handlers
		for sig, handler in self.old_handlers.items():
			signal.signal(sig, handler)

		# Call the handlers if necessary.
		for sig, handler in self.old_handlers.items():
			if handler:
				for args in self.received[sig]:
					handler(*args)

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import pickle
import logging
from collections import OrderedDict
import numpy
from . import Supplier
from ..sources import VanillaSource

logger = logging.getLogger(__name__)

###############################################################################
class PickleSupplier(Supplier):
	""" A supplier which parses out data from a Python pickle and exposes it as
		a numpy array. The pickled object is assumed to be a dictionary, with
		keys that map to numpy arrays.
	"""

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the supplier.
		"""
		return 'pickle'

	###########################################################################
	def __init__(self, source, *args, **kwargs):
		""" Creates a new Pickle supplier.

			# Arguments

			source: str. Filename of pickled file to load.
		"""
		super().__init__(*args, **kwargs)

		self.source = source
		self.data = None

	###########################################################################
	def _load(self):
		""" Loads the data (only if it hasn't already been loaded).
		"""
		if self.data is not None:
			return

		with open(self.source, 'rb') as fh:
			content = fh.read()

		try:
			data = pickle.loads(content)
		except UnicodeDecodeError:
			data = pickle.loads(content, encoding='latin1')
			logger.warning('We needed to explicitly set a "latin1" encoding '
				'to properly load the pickled data. This is probably because '
				'the pickled data was created in Python 2. You really should '
				'switch over to Python 3 in order to ensure future '
				'compatibility.')

		if not isinstance(data, (dict, OrderedDict)):
			raise ValueError('Pickled data must be a dictionary of numpy '
				'arrays.')

		result = {}
		for k, v in data.items():
			if not isinstance(k, str):
				logger.warning('Dictionary keys should be strings. Instead we '
					'received: %s. We are going to try to cast this to a '
					'and keep on going. This might break things. Please use '
					'string keys in the future.', k)
				k = str(k)

			if isinstance(v, (list, tuple)):
				v = numpy.array(v)
			elif not isinstance(v, numpy.ndarray):
				logger.warning('The value corresponding to pickled dictionary '
					'key "%s" is an unexpected type: %s. We are going to try '
					'and wrap this in a numpy array as see what happens. In '
					'future, be sure to store native numpy arrays as the '
					'dictionary keys, or at least try using Python lists or '
					'tuples.', k, type(v))
				v = numpy.array(v)

			if v.dtype.kind == 'O':
				logger.warning('The pickled numpy array we are trying to '
					'process for key "%s" has an underlying "object" data '
					'type. This could be because the array is ragged, or '
					'because the data itself is not numeric. We will try to '
					'use the data, but this could cause unexpected problems.',
					k)

			result[k] = VanillaSource(v)

		self.data = result

	###########################################################################
	def get_sources(self, sources=None):
		""" Returns all sources from this provider.
		"""
		self._load()

		if sources is None:
			sources = list(self.data.keys())
		elif not isinstance(sources, (list, tuple)):
			sources = [sources]

		for source in sources:
			if source not in self.data:
				raise KeyError(
					'Invalid data key: {}. Valid keys are: {}'.format(
						source, ', '.join(str(k) for k in self.data.keys())
				))

		return {k : self.data[k] for k in sources}

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import pytest
import numpy

from kur.utils import get_subclasses

from kur.engine import Engine
from kur.backend import Backend
from kur.containers import Container
from kur.model import Model
from kur.providers import BatchProvider
from kur.sources import VanillaSource
from kur.utils import can_import

################################################################################
def keras_mock(cls, backend, deps):
	""" Fakes a backend-looking object that can be used in place of a Keras
		backend. It is used to force certain variations of the Keras backend to
		get tested.
	"""
	result = lambda: cls(backend=backend)
	setattr(result, 'is_supported',
		lambda: cls.is_supported() and all(can_import(dep) for dep in deps))
	setattr(result, 'get_name',
		lambda: 'mock_{}_{}'.format(cls.get_name(), backend))
	return result

###############################################################################
def enumerate_backends():
	""" Enumerates all the backends.

		Instead of calling `Backend.get_all_backends()` directly, this layer of
		indirection allows us to "fake" particular backend variants.
	"""
	result = []

	all_backends = Backend.get_all_backends(supported_only=False)
	for backend in all_backends:
		if backend is Backend.get_backend_by_name('keras'):
			result.extend([
				keras_mock(backend, 'theano', ('theano', )),
				keras_mock(backend, 'theano', ('tensorflow', ))
			])
		else:
			result.append(backend)

	return result

###############################################################################
@pytest.fixture(
	params=enumerate_backends()
)
def a_backend(request):
	""" Fixture for obtaining a backend.
	"""
	cls = request.param
	if not cls.is_supported():
		pytest.xfail('Backend {} is not installed and cannot be tested.'
			.format(cls.get_name()))
	return cls()

###############################################################################
@pytest.fixture(
	params=get_subclasses(Engine, recursive=True)
)
def an_engine(request):
	""" Fixture for obtaining an engine.
	"""
	return request.param()

###############################################################################
def model_with_containers(backend, containers):
	""" Convenience function for creating a model instance from the toy model
		functions in this module (e.g., `simple_model()`).
	"""
	return Model(
		backend=backend,
		containers=[
			Container.create_container_from_data(container)
			for container in containers
		]
	)

###############################################################################
@pytest.fixture
def simple_model(a_backend):
	""" Returns a model that is extremely simple (one layer).
	"""
	return model_with_containers(
		backend=a_backend,
		containers=[
			{'input' : {'shape' : [10]}, 'name' : 'TEST_input'},
			{'dense' : 1, 'name' : 'TEST_output'},
		]
	)

###############################################################################
@pytest.fixture
def simple_data():
	""" Returns a small provider that can be used to train the `simple_model()`.
	"""
	return BatchProvider(
		sources={
			'TEST_input' : VanillaSource(numpy.random.uniform(size=(100, 10))),
			'TEST_output' : VanillaSource(numpy.random.uniform(size=(100, 1)))
		}
	)

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from ..utils import get_subclasses

###############################################################################
class Optimizer:
	""" Base class for all optimizers
	"""

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the optimizer.

			# Return value

			A lower-case string unique to this optimizer.
		"""
		return cls.__name__.lower()

	###########################################################################
	@staticmethod
	def get_all_optimizers():
		""" Returns an iterator to the names of all optimizers.
		"""
		for cls in get_subclasses(Optimizer):
			yield cls

	###########################################################################
	@staticmethod
	def get_optimizer_by_name(name):
		""" Finds a optimizer class with the given name.
		"""
		name = name.lower()
		for cls in Optimizer.get_all_optimizers():
			if cls.get_name() == name:
				return cls
		raise ValueError('No such optimizer with name "{}"'.format(name))

	###########################################################################
	def __init__(self):
		""" Creates a new optimizer.
		"""
		pass

	###########################################################################
	def get_optimizer(self, backend):
		""" Returns the optimizer that can be used by the implementation-
			specific model.
		"""
		raise NotImplementedError

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import importlib

def can_import(name, package=None):
	""" Returns True if the given module can be imported.

		# Arguments

		name: str. The name of the module.
		package: str. The name of the package, if `name` is a relative import.
			This is ignored for Python versions < 3.4.

		# Return value

		If importing the specified module should succeed, returns True;
		otherwise, returns False.
	"""
	try:
		importlib.util.find_spec
	except AttributeError:
		# Python < 3.4
		return importlib.find_loader(		# pylint: disable=deprecated-method
			name
		) is not None
	else:
		# Python >= 3.4
		return importlib.util.find_spec(name, package=package) is not None

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
from ..utils import get_subclasses

logger = logging.getLogger(__name__)

###############################################################################
class Backend:
	""" Base class for all tensor operation backends. Examples of backends are
		Keras, Theano, TensorFlow, etc.
	"""

	###########################################################################
	def __init__(self, variant=None):
		""" Create a new backend.

			Part of this call should be to ensure that all the necessary
			modules/libraries are available to use this backend. If this
			backend cannot be used, an exception should be raised. Otherwise,
			any appropriate configuration should take place here.

			# Arguments

			variant: str, list/tuple of str, or None (default: None). The
			variants to label this backend with. This allows for more
			streamlined and flexible modification of backend behavior at the
			layer level. These are simply tags that software components can
			query and modify their behavior in response, without requiring an
			entirely new backend.
		"""
		if not self.is_supported():
			logger.warning('Backend claims to not be supported. We will try '
				'to use it anyway.')

		if variant is not None:
			if isinstance(variant, str):
				variant = {variant}
		else:
			variant = set()

		self.variant = set(variant)

		logger.info('Creating backend: %s', self.get_name())
		logger.info('Backend variants: %s',
			'none' if not self.variant else ', '.join(
				str(x) for x in sorted(self.variant)
			)
		)

	###########################################################################
	def has_variant(self, variant):
		""" Checks if a particular variant is enabled.
		"""
		return variant in self.variant

	###########################################################################
	@classmethod
	def is_supported(cls):
		""" Returns True if this backend can be used.

			# Return value

			If the backend appears to be useable, returns True. Otherwise,
			returns False.

			Note that if this returns False, then the backend should definitely
			not be able to be used. However, just because it returns True
			doesn't mean it will work (e.g., Theano with `force_device=true`,
			but without the requested device installed).
		"""
		raise NotImplementedError

	###########################################################################
	@staticmethod
	def from_specification(spec):
		""" Creates a new backend from the specification.

			# Arguments

			spec: str or dict. The backend specification. If processed from
				a standard job specification, it is everything under the
				`global.backend` section.

			# Return value

			Backend instance

			# Usage

			If no "global" section exists, or if no "backend" section exists in
			the "global" section, this will try to instantiate any supported
			backend installed on the system.

			Instantiate a specific backend with default parameters:
			```yaml
			global:
			  backend: BACKEND
			```
			or like this:
			```yaml
			global:
			  backend:
			    name: BACKEND
			```

			Instantiate a specific backend with additional keywords (the
			keywords must be supported by the Backend implementation):
			```yaml
			global:
			  backend:
			    name: BACKEND
				params:
				  PARAM1: VALUE1
				  PARAM2: VALUE2
				  ...
			```

			Instatiate any backend with additional keywords (the keywords must
			be supported by the Backend implementation):
			```yaml
			global:
			  backend:
				params:
				  PARAM1: VALUE1
				  PARAM2: VALUE2
				  ...
			```
		"""
		if spec is None:
			all_supported = list(Backend.get_all_backends(supported_only=True))
			if not all_supported:
				raise ValueError('No supported backends available.')
			target = all_supported[0]
			params = {}
		elif isinstance(spec, str):
			target = Backend.get_backend_by_name(spec)
			params = {}
		elif isinstance(spec, dict):
			variant = spec.get('variant')
			params = spec.get('params', {})
			params = dict(params)
			params['variant'] = variant

			if 'name' in spec:
				target = Backend.get_backend_by_name(
					spec['name']
				)
			else:
				all_supported = list(Backend.get_all_backends(
					supported_only=True))
				if not all_supported:
					raise ValueError('No supported backends available.')
				target = all_supported[0]
		else:
			raise ValueError(
				'Unexpected backend specification: {}'.format(spec))

		logger.debug('Using backend: %s', target.get_name())
		result = target(**params)
		return result

	###########################################################################
	@staticmethod
	def get_backend_by_name(name):
		""" Factory method for creating new backends.

			# Arguments

			name: str. The name of the backend to instantiate. This is given by
				the backend's `get_name()` method.
			kwargs: dict. The arguments to pass to the backend's constructor.

			# Return value

			If the named backend is found, this returns a Backend instance;
			otherwise, a ValueError is raised.
		"""
		name = name.lower()
		for cls in Backend.get_all_backends(supported_only=False):
			if cls.get_name() == name:
				return cls
		raise ValueError('No such backend: {}'.format(name))

	###########################################################################
	@staticmethod
	def get_all_backends(supported_only=False):
		""" Factory method for creating backend which is supported on the host
			platform.

			# Arguments

			kwargs: dict. The arguments to pass to the backend's constructor.

			# Return value

			If the named backend is found, this returns a Backend instance;
			otherwise, a ValueError is raised.
		"""
		for cls in get_subclasses(Backend):
			if supported_only:
				if cls.is_supported():
					yield cls
			else:
				yield cls

	###########################################################################
	@staticmethod
	def get_any_supported_backend():
		""" Finds a backend class that claims to be supported on this system.
		"""
		for cls in Backend.get_all_backends(supported_only=True):
			return cls
		raise ValueError('No supported backends found.')

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Return a lower-case string naming this backend.
		"""
		return cls.__name__.lower()

	###########################################################################
	def save(self, model, filename):
		""" Saves the model weights to the given filename.

			# Arguments

			model: Model instance. The model whose weights should be saved.
			filename: str. The filename to write the weights to.

			# Notes

			The file format is backend-specific. There is no guarantee of
			compatability between different backends.

			# Return value

			None
		"""
		raise NotImplementedError

	###########################################################################
	def restore(self, model, filename):
		""" Load the model weights from the given filename.

			# Arguments

			model: Model instance. The model whose weights should be restored.
			filename: str. The filename to read the weights from.

			# Notes

			The file format is backend-specific. There is no guarantee of
			compatability between different backends.

			# Return value

			None
		"""
		raise NotImplementedError

	###########################################################################
	def connect(self, inputs, target):
		""" Applies a tensor operation to a set of input tensors.

			# Arguments

			inputs: list. A list of input tensors.
			target: object. The tensor operation to apply.

			# Notes

			The exact underlying types of the `inputs` and `target` tensors
			depend on the backend implementation.

			# Return value

			A new tensor object (backend-specific) resulting from the applied
			operation.
		"""
		raise NotImplementedError

	###########################################################################
	def compile(self, model, loss=None, optimizer=None):
		""" Creates an implementation-specific representation of the
			instantiated model.

			# Arguments

			model: Model instance. The built model to create a representation
				for.
			loss: Loss instance, list, dict, or None. The loss function
				(objective function) to use for training. If None, no loss
				functions are used. If a list, it should be a list of Loss
				instances, one for each model output; if the list is length 1,
				then the same loss function is used for each model output. If
				this is a dict, it should map from model output names to Loss
				instances to apply to that output.
			optimizer: Optimizer instance or None. The loss function to use for
				training. If None, no optimizer is used.

			# Return value

			An implementation-specific object ("model") that can be used for
			training or evaluation.

			# Notes

			Implementations may need to modify the Model. If this is the case,
			ExtensionState should be used to protect the model from extensions
			that are left stuck on after compiling.
		"""
		raise NotImplementedError

	###########################################################################
	def train(self, model, data, compiled):
		""" Fits the given model on a batch of data.

			# Arguments

			model: Model instance. The model to train with.
			data: dict or list. The data to train a batch on.
			compiled: object. The data package returned by `compile`. Its type
				and content are backend-specific.

			# Return value

			A dictionary of loss values, whose keys are the names of the
			respective outputs. If per-output loss is not available, then a
			single, global loss value can be returned instead.
		"""
		raise NotImplementedError

	###########################################################################
	def test(self, model, data, compiled):
		""" Calculates the model loss on a batch of data.

			# Arguments

			model: Model instance. The model to evaluate with.
			data: dict or list. The batch of data to test on.
			compiled: object. The data package returned by `compile`. Its type
				and content are backend-specific.

			# Return value

			A dictionary of loss values, whose keys are the names of the
			respective outputs. If per-output loss is not available, then a
			single, global loss value can be returned instead.
		"""
		raise NotImplementedError

	###########################################################################
	def evaluate(self, model, data, compiled):
		""" Evaluates the model on a batch ofdata.
		"""
		raise NotImplementedError

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import contextlib
import io
import re
import os
import sys
import logging
import numpy
from . import Backend
from ..containers import Layer
from ..loss import Loss
from ..model import ExtensionState
from ..utils import can_import, EnvironmentalVariable, redirect_stderr, idx

logger = logging.getLogger(__name__)

###############################################################################
class KerasBackend(Backend):
	""" A Keras backend.

		# Dependencies

		- keras
		- theano OR tensorflow
		- h5py
	"""

	###########################################################################
	@classmethod
	def is_supported(cls):
		""" Returns True if this backend can be used.
		"""
		return can_import('keras') and (
			can_import('theano') or can_import('tensorflow')
		)

	###########################################################################
	def __init__(self, backend=None, *args, **kwargs):
		""" Creates a new Keras backend.

			As per the base class documentation, we should do all necessary
			Keras-related initialization here, including checking obvious
			things like "Is Keras installed?" or "Is a backend installed?"

			# Arguments

			backend: str or None (default: None). The Keras backend to use
			(either "theano" or "tensorflow"). None uses the system default.
		"""

		super().__init__(*args, **kwargs)

		if backend is not None:
			logger.info('The %s backend for Keras has been requested.')

			if 'keras' in sys.modules:
				import keras.backend as K		# pylint: disable=import-error
				if K.backend() != backend:
					logger.warning('Keras was already imported by the time '
						'the Kur backend was instantiated. Kur was asked to '
						'use Keras %s backend, but Keras is already using %s. '
						'We cannot change the Keras backend at this point, so '
						'we will try to work with the currently loaded '
						'backend. In the future, try to let Kur manage '
						'importing Keras.', backend, K.backend())

			for dep in {'theano' : ['theano'], 'tensorflow' : ['tensorflow']}[backend]:
				if not can_import(dep):
					logger.warning('The Keras backend was asked to use the %s '
						'backend, but %s does not appear to be installed. You '
						'will likely get an error about this soon.',
						backend, dep)

		else:
			logger.info('No particular backend for Keras has been requested.')
			if can_import('theano') and can_import('tensorflow'):
				logger.debug('Using the system-default Keras backend.')
			elif can_import('theano'):
				backend = 'theano'
				logger.debug('Only the Theano backend for Keras is installed, '
					'so we will try to use it.')
			elif can_import('tensorflow'):
				backend = 'tensorflow'
				logger.debug('Only the TensorFlow backend for Keras is '
					'installed, so we will try to use it.')
			else:
				logger.warning('No supported Keras backend seems to be '
					'installed. You will probably get an error about this '
					'shortly.')

		# Make sure Keras is loaded.
		# Now, Keras always prints out a "Using {Theano|TensorFlow} backend."
		# statement that is frankly unbecoming. So we'll just gobble it up here.
		x = io.StringIO()
		with redirect_stderr(x):

			with EnvironmentalVariable(KERAS_BACKEND=backend):
				import keras	# pylint: disable=import-error,unused-variable
				import keras.backend as K		# pylint: disable=import-error
				logger.debug('Keras is loaded. The backend is: %s',
					K.backend())

		# And now we can set the dimension ordering.
		keras.backend.set_image_dim_ordering('tf')

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the backend class.

			This is used by the base class's factory method.
		"""
		return 'keras'

	###########################################################################
	def connect(self, inputs, target):
		""" Use the Keras functional API to connect to layers

			# Notes:

			- You will need input placeholders in place before doing this,
			  otherwise Keras's shape-checking will fail.
		"""
		if isinstance(inputs, Layer):
			inputs = [inputs]
		return target(inputs)

	###########################################################################
	def save(self, model, filename):
		""" Saves the model weights to the given filename.
		"""
		import keras.models as M				# pylint: disable=import-error

		keras_model = M.Model(
			input=[node.value for node in model.inputs.values()],
			output=[node.value for node in model.outputs.values()]
		)

		path = os.path.expanduser(os.path.expandvars(filename))
		if os.path.exists(path):
			if not os.path.isdir(path):
				raise ValueError('Target weight exists, but it is not a '
					'directory. Kur expected a directory that it can work '
					'with. Please move or delete the existing path: {}'
					.format(path))
		else:
			os.makedirs(path, exist_ok=True)

		layers = keras_model.flattened_layers \
			if hasattr(keras_model, 'flattened_layers') else keras_model.layers
		for layer in layers:
			layer_name = layer.name

			symbolic_weights = layer.weights
			weight_names, weight_values = \
				self._get_weight_names_and_values_from_symbolic(
					symbolic_weights
				)

			for name, val in zip(weight_names, weight_values):
				target = os.path.join(
					path,
					'{}+{}.kur'.format(layer_name, name)
				)
				idx.save(target, val)

	###########################################################################
	def _get_weight_names_and_values_from_symbolic(self, symbolic_weights):
		import keras.backend as K				# pylint: disable=import-error
		weight_values = K.batch_get_value(symbolic_weights)
		weight_names = [
			(
				str(w.name) if hasattr(w, 'name') and w.name \
					else 'param_{}'.format(i)
			)
			for i, (w, val) in enumerate(
				zip(symbolic_weights, weight_values)
			)
		]
		return weight_names, weight_values

	###########################################################################
	def restore(self, model, filename):
		""" Load the model weights from the given filename.
		"""
		import keras.models as M				# pylint: disable=import-error
		import keras.backend as K				# pylint: disable=import-error

		keras_model = M.Model(
			input=[node.value for node in model.inputs.values()],
			output=[node.value for node in model.outputs.values()]
		)

		path = os.path.expanduser(os.path.expandvars(filename))
		if os.path.exists(path):
			if not os.path.isdir(path):
				raise ValueError('Target weight exists, but it is not a '
					'directory. Kur expected a directory that it can work '
					'with. Please move or delete the existing path: {}'
					.format(path))
		else:
			raise ValueError('Target weight directory does not exist: {}'
				.format(path))

		layers = keras_model.flattened_layers \
			if hasattr(keras_model, 'flattened_layers') else keras_model.layers

		# Get a map from "layer name" to "layer instance" in the current model.
		index = {}
		for layer in layers:
			if layer.name:
				index.setdefault(layer.name, []).append(layer)

		# Enumerate all of the saved tensors, organized like this:
		# tensors = {
		#	'layer_1_name' : {
		#		'weight_1_name'  : '/path/to/file',
		#		...,
		#	},
		#	...
		# }
		tensors = self.enumerate_saved_tensors(path)

		# We want to put (symbolic_weights, weight_values) tuples in this.
		weight_value_tuples = []

		# Loop over the available weights.
		for layer_name, weights in tensors.items():

			# Load the weights.
			# This maps weight names to numpy arrays.
			weights = {k : idx.load(v) for k, v in weights.items()}

			# Now assign all of the weights to their corresponding symbolic
			# weights. Loop over all layers which use this name.
			for layer in index.get(layer_name, []):

				# Get the symbolic weights.
				symbolic_weights = layer.weights
				if len(weights) != len(symbolic_weights):
					raise ValueError('Layer "%s" expected %d weights, but we '
						'found %d on disk.', layer_name, len(symbolic_weights),
						len(weights))

				# Get the associated names (so we know what order to assign the
				# weights in.
				weight_names, _ = \
					self._get_weight_names_and_values_from_symbolic(
						symbolic_weights
					)
				for i, name in enumerate(weight_names):
					weight_value_tuples.append((symbolic_weights[i], weights[name]))

		# Assign all the weights in one batch (for efficiency).
		K.batch_set_value(weight_value_tuples)

	###########################################################################
	def enumerate_saved_tensors(self, path):
		""" Enumerates saved tensors (weights).
		"""
		result = {}

		regex = re.compile(r'^(?P<layer>.*?)\+(?P<weight>.*?)\.kur$')
		for dirpath, dirnames, filenames in os.walk(path): # pylint: disable=unused-variable
			for filename in filenames:
				match = regex.match(filename)
				if match is None:
					continue
				filename = os.path.join(dirpath, filename)
				layer, weight = match.groups()
				if layer not in result:
					result[layer] = {}
				if weight in result[layer]:
					logger.warning('Tensor weights have already been loaded '
						'for layer="%s", tensor="%s" from file: %s. We will '
						'skip this new file we just found: %s.', layer, weight,
						result[layer][weight], filename)
					continue
				result[layer][weight] = filename

		return result

	###########################################################################
	def compile(self, model, loss=None, optimizer=None):
		""" Returns the Keras model instance.
		"""

		with ExtensionState(model):

			if loss is not None:
				loss, alias = self._apply_loss(model, loss)
				rev = {v : k for k, v in alias.items()}
			else:
				loss, alias, rev = None, {}, {}

			import keras.models as M			# pylint: disable=import-error
			logger.debug('Instantiating a Keras model.')
			result = M.Model(
				input=[node.value for node in model.inputs.values()],
				output=[node.value for node in model.outputs.values()]
			)

			if logger.isEnabledFor(logging.DEBUG):
				x = io.StringIO()
				with contextlib.redirect_stdout(x):
					result.summary()
				for line in x.getvalue().split('\n'):
					logger.debug(line)

			if loss is not None:# and optimizer is not None:
				logger.debug('Starting to compile the Keras model.')
				result.compile(
					loss={name : func.get_loss(self)
						for name, func in loss.items()},
					optimizer=optimizer.get_optimizer(self) \
						if optimizer is not None else None,
					loss_weights={name : func.get_weight()
						for name, func in loss.items()}
				)

			return {'model' : result, 'alias' : alias, 'rev_alias' : rev}

	###########################################################################
	def _apply_loss(self, model, loss=None):	# pylint: disable=no-self-use
		""" Applies the loss functions to the model.

			# Arguments

			# Return value
		"""
		# How many outputs are in this model?
		num_outputs = len(model.outputs)

		if isinstance(loss, Loss):
			loss = [loss]

		if isinstance(loss, (list, tuple)):
			if len(loss) == 1:
				while len(loss) < num_outputs:
					loss.append(loss[0])
			elif len(loss) != num_outputs:
				raise ValueError('Wrong number of loss functions specified. '
					'There must be as many loss functions as model outputs, or '
					'a single loss function that can be replicated for each '
					'model output. In this case, there were {} loss functions '
					'given, but the model has {} outputs.'
						.format(len(loss), num_outputs))

			temp = {}
			for name, output_loss in zip(model.outputs, loss):
				temp[name] = output_loss
			loss = temp				# pylint: disable=redefined-variable-type

		elif not isinstance(loss, dict):
			raise ValueError('Unexpected form for the loss function. Expected '
				'a single loss function, a list of loss functions, or a '
				'dictionary mapping model outputs to loss functions. Instead '
				'we received this: {}'.format(loss))

		supplied_loss = set(loss.keys())
		required_loss = set(model.outputs.keys())

		for name in supplied_loss - required_loss:
			logger.warning('Loss function was supplied for "%s", but no such '
				'output exists in the model. Maybe you meant one of these: %s. '
				'Supplied loss functions: %s. Required loss functions: %s.',
				name,
				', '.join(x for x in required_loss - supplied_loss),
				', '.join(supplied_loss),
				', '.join(required_loss)
			)
		missing_loss = False
		for name in required_loss - supplied_loss:
			logger.error('Loss function is needed for model output "%s", but '
				'no such loss function was supplied. Maybe you meant one of '
				'these: %s. Supplied loss functions: %s. Required loss '
				'functions: %s.',
				name,
				', '.join(x for x in supplied_loss - required_loss),
				', '.join(supplied_loss),
				', '.join(required_loss)
			)
			missing_loss = True
		if missing_loss:
			raise ValueError('One or more model outputs did not have loss '
				'functions supplied. Loss functions are needed for these '
				'model outputs: {}'.format(
					', '.join(name for name in required_loss - supplied_loss)
			))

		# Once the model extension retracts, the alias keys still exist as the
		# Model inputs/outputs, and they will match data source names.
		# But the Keras model has inputs/outputs named after `alias[name]`.
		alias = {name : func.modify(model, name)
			for name, func in loss.items()}

		return loss, alias

	###########################################################################
	def train(self, model, data, compiled):
		""" Fits the given model on a batch of data.
		"""
		metrics = compiled['model'].train_on_batch(
			{compiled['alias'].get(name, name) : data[name]
				for name in model.inputs},
			{compiled['alias'].get(name, name) : data[name]
				for name in model.outputs}
		)

		return KerasBackend._convert_metrics(
			metrics,
			compiled['model'].metrics_names,
			compiled['rev_alias'],
			model.outputs
		)

	###########################################################################
	def test(self, model, data, compiled):
		""" Calculates the model loss on a batch of data.
		"""
		metrics = compiled['model'].test_on_batch(
			{compiled['alias'].get(name, name) : data[name]
				for name in model.inputs},
			{compiled['alias'].get(name, name) : data[name]
				for name in model.outputs}
		)

		return KerasBackend._convert_metrics(
			metrics,
			compiled['model'].metrics_names,
			compiled['rev_alias'],
			model.outputs
		)

	###########################################################################
	@staticmethod
	def _convert_metrics(metrics, metrics_names, rev_alias, outputs):
		""" Formats the Keras metrics properly for use by Kur.
		"""
		if not isinstance(metrics, list):
			metrics = [metrics]
		un_numpy = lambda x: x.tolist() if isinstance(x, numpy.ndarray) else x
		metrics = {k : un_numpy(v) for k, v in zip(metrics_names, metrics)}

		loss = {}
		for k, v in metrics.items():
			match = re.match(r'^(.*)_loss$', k)
			if match:
				k = match.group(1)
				k = rev_alias.get(k, k)
				loss[k] = float(v)
		if not loss:
			for k in outputs:
				break
			loss[k] = metrics['loss']

		return loss

	###########################################################################
	def evaluate(self, model, data, compiled):
		""" Evaluates the model on a batch of data.
		"""
		# Returns an array of model outputs, with one entry per branch.
		results = compiled['model'].predict_on_batch(
			{compiled['alias'].get(name, name) : data[name]
				for name in model.inputs}
		)

		if len(model.outputs) == 1:
			results = [results]

		return {name : result for name, result in zip(model.outputs, results)}

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from . import Layer, ParsingError

###############################################################################
class Dense(Layer):						# pylint: disable=too-few-public-methods
	""" A fully-connected layer.
	"""

	DEFAULT_AUTO_FLATTEN = False

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Creates a new dense layer.
		"""
		super().__init__(*args, **kwargs)
		self.size = None
		self.auto_flatten = None

	###########################################################################
	def _parse(self, engine):
		""" Parse the layer.
		"""

		super()._parse(engine)

		if isinstance(self.args, dict):
			self.size = engine.evaluate(self.args['size'], recursive=True)
			self.auto_flatten = engine.evaluate(
				self.args.get('flatten', Dense.DEFAULT_AUTO_FLATTEN),
				recursive=True
			)
		elif isinstance(self.args, list):
			self.size = engine.evaluate(self.args, recursive=True)
			self.auto_flatten = Dense.DEFAULT_AUTO_FLATTEN
		else:
			self.size = self.args
			self.auto_flatten = Dense.DEFAULT_AUTO_FLATTEN

		if not isinstance(self.size, (tuple, list)):
			self.size = [self.size]

		try:
			for i, v in enumerate(self.size):
				self.size[i] = int(v)
		except ValueError:
			raise ParsingError('Key "size" in Dense layer must be an integer '
				'or a list of integers. Received: {}'.format(self.size))

		if not isinstance(self.auto_flatten, bool):
			raise ParsingError('"auto_flatten" key in Dense layer must be '
				'boolean. Received: {}'.format(self.auto_flatten))

	###########################################################################
	def _build(self, backend):
		""" Create the backend-specific placeholder.
		"""
		if backend.get_name() == 'keras':

			import keras.layers as L			# pylint: disable=import-error

			if self.auto_flatten:
				yield L.Flatten()

			for v in self.size[:-1]:
				yield L.Dense(output_dim=v)

			yield L.Dense(
				output_dim=self.size[-1],
				name=self.name
			)

		else:
			raise ValueError(
				'Unknown or unsupported backend: {}'.format(backend))

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from . import Layer, ParsingError

###############################################################################
class Parallel(Layer):					# pylint: disable=too-few-public-methods
	""" A container for applying tensor operations to each element in a list.
		This is effectively a map operation, like Theano's `scan` or Keras's
		`TimeDistributed`.

		# Example

		```
		parallel:
		  apply:
		    - dense:
		        size: 512
			- convolution:
			    kernels: 16
				...
		```
	"""

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Create a new parallel container.
		"""
		super().__init__(*args, **kwargs)

	###########################################################################
	def _parse(self, engine):
		""" Parse the child containers
		"""

		# Always call the parent.
		super()._parse(engine)

		# Parse self
		if 'apply' not in self.args:
			raise ParsingError('Missing "apply" key in parallel container.')
		target = engine.evaluate(self.args['apply'])

		# Parse children
		for child in target:
			self.new_child_from_data(child).parse(engine)

	###########################################################################
	def _build(self, backend):
		""" Instantiate the container.
		"""
		for child in self.children:
			for layer in child.build(backend):

				if backend.get_name() == 'keras':

					import keras.layers as L	# pylint: disable=import-error
					yield L.TimeDistributed(
						layer,
						name='{}_{}'.format(self.name, child.name)
					)

				else:
					raise ValueError(
						'Unknown or unsupported backend: {}'.format(backend)
					)

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import sys
import argparse
import logging
from .utils import logcolor
from .parsing import Specification
from .engine import JinjaEngine

logger = logging.getLogger(__name__)

###############################################################################
def parse_specification(filename, engine):
	""" Parses a specification file.

		# Arguments

		filename: str. The path to the specification file to load.

		# Return value

		Specification instance
	"""
	spec = Specification(filename, engine)
	spec.parse()
	return spec

###############################################################################
def train(args):
	""" Trains a model.
	"""
	spec = parse_specification(args.specification, args.engine)
	func = spec.get_training_function()
	func()

###############################################################################
def test(args):
	""" Tests a model.
	"""
	spec = parse_specification(args.specification, args.engine)
	func = spec.get_testing_function()
	func()

###############################################################################
def evaluate(args):
	""" Evaluates a model.
	"""
	spec = parse_specification(args.specification, args.engine)
	func = spec.get_evaluation_function()
	func()

###############################################################################
def build(args):
	""" Builds a model.
	"""
	spec = parse_specification(args.specification, args.engine)
	spec.get_model()

	if args.compile == 'none':
		return
	elif args.compile == 'train':
		target = spec.get_trainer(with_optimizer=True)
	elif args.compile == 'test':
		target = spec.get_trainer(with_optimizer=False)
	elif args.compile == 'evaluate':
		target = spec.get_evaluator() # pylint: disable=redefined-variable-type
	else:
		logger.error('Unhandled compilation target: %s. This is a bug.',
			args.compile)
		return

	target.compile()

###############################################################################
def parse_args():
	""" Constructs an argument parser and returns the parsed arguments.
	"""
	parser = argparse.ArgumentParser(
		description='Descriptive deep learning')
	parser.add_argument('--no-color', action='store_true',
		help='Disable colorful logging.')
	parser.add_argument('-v', '--verbose', default=0, action='count',
		help='Increase verbosity. Can be specified twice for debug-level '
			'output.')

	subparsers = parser.add_subparsers(dest='cmd', help='Sub-command help.')
	subparsers.required = True

	subparser = subparsers.add_parser('train', help='Trains a model.')
	subparser.add_argument('specification', nargs='?',
		help='The specification file to use.')
	subparser.set_defaults(func=train)

	subparser = subparsers.add_parser('test', help='Tests a model.')
	subparser.add_argument('specification', nargs='?',
		help='The specification file to use.')
	subparser.set_defaults(func=test)

	subparser = subparsers.add_parser('evaluate', help='Evaluates a model.')
	subparser.add_argument('specification', nargs='?',
		help='The specification file to use.')
	subparser.set_defaults(func=evaluate)

	subparser = subparsers.add_parser('build',
		help='Tries to build a model. This is useful for debugging a model.')
	subparser.add_argument('specification', nargs='?',
		help='The specification file to use.')
	subparser.add_argument('-c', '--compile',
		choices=['none', 'train', 'test', 'evaluate'], default='none',
		help='Also try to compile the specified variation of the model.')
	subparser.set_defaults(func=build)

	return parser.parse_args()

###############################################################################
def main():
	""" Entry point for the Kur command-line script.
	"""
	args = parse_args()

	loglevel = {
		0 : logging.WARNING,
		1 : logging.INFO,
		2 : logging.DEBUG
	}
	config = logging.basicConfig if args.no_color else logcolor.basicConfig
	config(
		level=loglevel.get(args.verbose, logging.DEBUG),
		format='{color}[%(levelname)s %(asctime)s %(name)s:%(lineno)s]{reset} '
			'%(message)s'.format(
				color='' if args.no_color else '$COLOR',
				reset='' if args.no_color else '$RESET'
			)
	)
	logging.captureWarnings(True)

	engine = JinjaEngine()
	setattr(args, 'engine', engine)

	sys.exit(args.func(args) or 0)

###############################################################################
if __name__ == '__main__':
	main()

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from . import Loss

###############################################################################
class CategoricalCrossentropy(Loss):
	""" Categorical crossentropy loss, used for 1-of-N classification.
	"""

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the loss function.
		"""
		return 'categorical_crossentropy'

	###########################################################################
	def get_loss(self, backend):

		if backend.get_name() == 'keras':
			return 'categorical_crossentropy'
		else:
			raise ValueError('Unsupported backend "{}" for loss function "{}"'
				.format(backend.get_name(), self.get_name()))

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import tarfile
import os
import re
import pickle

import numpy

from ..utils import download_file
from . import Supplier
from ..sources import VanillaSource

###############################################################################
class CifarSupplier(Supplier):
	""" A supplier which supplies MNIST image/label pairs. These are downloaded
		from the internet, verified, and parsed as IDX files.
	"""

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the supplier.
		"""
		return 'cifar'

	###########################################################################
	def __init__(self, url=None, sha256=None, local=None, parts=None,
		*args, **kwargs):
		""" Creates a new CIFAR supplier.

			# Arguments

		"""
		super().__init__(*args, **kwargs)

		path = CifarSupplier._get_filename({
			'url' : url,
			'sha256' : sha256,
			'local' : local
		})
		images, labels = CifarSupplier._load_parts(path, parts)

		self.data = {
			'images' : VanillaSource(CifarSupplier._normalize(images)),
			'labels' : VanillaSource(CifarSupplier._onehot(labels))
		}

	###########################################################################
	@staticmethod
	def _load_parts(path, parts):

		if parts is not None:
			if not isinstance(parts, (list, tuple)):
				parts = [parts]

		result = {}

		with tarfile.open(path, 'r') as tar:
			for member in tar.getmembers():
				_, filename = os.path.split(member.name)

				match = re.match('^data_batch_([0-9]+)$', filename)
				if match:
					val = match.group(1)
				elif filename == 'test_batch':
					val = 'test'
				else:
					continue

				i = val # shut pylint up
				if parts is not None:
					for i in parts:
						if str(i) == val:
							break
					else:
						continue

				if i in result:
					raise ValueError('Too many matches in extracted CIFAR '
						'data for: {}'.format(i))

				stream = tar.extractfile(member)
				if stream is None:
					continue

				content = stream.read()
				data = pickle.loads(content, encoding='latin1')
				result[i] = data

		if parts is None:
			return list(result.values())

		if len(result) != len(parts):
			raise ValueError('Failed to find all pieces of the extract CIFAR '
				'data: {}'.format(parts))

		return (
			numpy.concatenate([result[i]['data'] for i in parts]),
			numpy.concatenate([result[i]['labels'] for i in parts])
		)

	###########################################################################
	@staticmethod
	def _onehot(source):
		onehot = numpy.zeros((source.shape[0], 10))
		for i, row in enumerate(source):
			onehot[i][row] = 1
		return onehot

	###########################################################################
	@staticmethod
	def _normalize(source):
		# Numpy won't automatically promote the uint8 fields to float32.
		data = source.astype(numpy.float32)
		data -= data.mean(axis=0)
		data /= 255
		data = data.reshape((-1, 3, 32, 32))
		data = numpy.transpose(data, axes=(0, 2, 3, 1))
		return data

	###########################################################################
	@staticmethod
	def _get_filename(target):
		""" Returns the filename associated with a particular target.

			# Arguments

			target: str or dict. The target specification. For locally-stored
				files, it can be a string (path to file) or a dictionary with
				key 'local' that contains the file path. For network files,
				it is a dictionary with 'url' (source URL); it may also
				optionally contain 'sha256' (SHA256 checksum) and 'local'
				(local storage directory for the file).

			# Return value

			String to the file's locally stored path. May not exist.
		"""

		if isinstance(target, str):
			return target
		elif isinstance(target, dict):
			if 'url' in target:
				return download_file(
					url=target['url'],
					sha256=target.get('sha256'),
					target_dir=target.get('path')
				)
			elif 'local' in target:
				return target['path']
			else:
				raise ValueError('Expected either "url" (for downloading data '
					'sources) or "local" (for locally-stored sources), but '
					'neither key was found in the CIFAR specification: {}'
					.format(target))
		else:
			raise ValueError('Unexpected data type for CIFAR target: {}'
				.format(target))

	###########################################################################
	def get_sources(self, sources=None):
		""" Returns all sources from this provider.
		"""

		if sources is None:
			sources = list(self.data.keys())
		elif not isinstance(sources, (list, tuple)):
			sources = [sources]

		for source in sources:
			if source not in self.data:
				raise KeyError(
					'Invalid data key: {}. Valid keys are: {}'.format(
						source, ', '.join(str(k) for k in self.data.keys())
				))

		return {k : self.data[k] for k in sources}

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import os
import logging
from .logger import Logger
from ..utils import idx

logger = logging.getLogger(__name__)

###############################################################################
class BinaryLogger(Logger):
	""" A class for storing log data in a fast binary format that can be
		quickly appended to or randomly seeked.
	"""

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the logger.
		"""
		return 'binary'

	###########################################################################
	def __init__(self, path, **kwargs):
		""" Creates a new binary logger.

			# Arguments

			path: str. The path to create the log at.

			# Notes

			- The BinaryLogger stores its log information in a directory. Thus,
			  `path` should be a directory.
		"""
		super().__init__(**kwargs)

		self.path = path
		self.prepare()

	###########################################################################
	def prepare(self):
		""" Prepares the logger by loading historical information.
		"""
		path = os.path.expanduser(os.path.expandvars(self.path))

		if os.path.exists(path):
			if os.path.isdir(path):
				logger.info('Loading log data: %s', path)

				training_loss = BinaryLogger.load_column(
					path, 'training_loss_total')
				if training_loss is None:
					self.best_training_loss = None
					self.num_epochs = None
				else:
					self.best_training_loss = training_loss.min()
					self.num_epochs = len(training_loss)

				validation_loss = BinaryLogger.load_column(
					path, 'validation_loss_total')
				if validation_loss is None:
					self.best_validation_loss = None
				else:
					self.best_validation_loss = validation_loss.min()

			else:
				raise ValueError('Binary logger stores its information in a '
					'directory. The supplied log path already exists, but it '
					'is a file: {}'.format(path))

		else:
			logger.info('Log does not exist. Creating path: %s', path)
			os.makedirs(path, exist_ok=True)

			self.best_training_loss = None
			self.num_epochs = None
			self.best_validation_loss = None

	###########################################################################
	def get_best_training_loss(self):
		""" Returns the best historical training loss.
		"""
		return self.best_training_loss

	###########################################################################
	def get_best_validation_loss(self):
		""" Returns the best historical validation loss.
		"""
		return self.best_validation_loss

	###########################################################################
	def get_number_of_epochs(self):
		""" Returns the number of epochs this model has historically completed.
		"""
		return self.num_epochs

	###########################################################################
	def process(self, data, data_type, tag=None):
		""" Processes training statistics.
		"""
		path = os.path.expanduser(os.path.expandvars(self.path))
		for k, v in data.items():
			column = '{}_{}_{}'.format(data_type, tag, k)

			filename = os.path.join(path, column)

			logger.debug('Adding data to binary column: %s', column)
			idx.save(filename, v, append=True)

	###########################################################################
	@staticmethod
	def load_column(path, column):
		""" Loads logged information from disk.

			# Arguments

			path: str. The path to the log directory.
			column: str. The name of the statistic to load.

			# Return value

			If the statistic specified by `column` exists in the log directory
			`path`, then a numpy array containing the stored data is returned.
			Otherwise, None is returned.
		"""
		logger.debug('Loading binary column: %s', column)
		path = os.path.expanduser(os.path.expandvars(path))
		if not os.path.isdir(path):
			logger.debug('No such log path exists: %s', path)
			return None

		filename = os.path.join(path, column)
		if not os.path.isfile(filename):
			logger.debug('No such log column exists: %s', filename)
			return None

		return idx.load(filename)

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
from collections import deque
import numpy
from ..utils import get_subclasses, Timer, CriticalSection

logger = logging.getLogger(__name__)

###############################################################################
class Logger:
	""" Base class for loggers.

		Loggers are special callbacks intended to process training statistics
		(such as loss).
	"""

	###########################################################################
	@staticmethod
	def from_specification(spec):
		""" Creates a new evaluation hook from a specification.

			# Arguments

			spec: str or dict. The specification object that was given.

			# Return value

			New EvaluationHook instance.
		"""
		if isinstance(spec, str):
			cls = Logger.DEFAULT_LOGGER
			params = {'path' : spec}
		elif isinstance(spec, dict):
			spec = dict(spec)
			name = spec.pop('name', None)
			if name is None:
				cls = Logger.DEFAULT_LOGGER
			else:
				cls = Logger.get_logger_by_name(name)
			params = spec
		else:
			raise ValueError('Expected the logger specification to be a string '
				'or dictionary. We got this instead: {}'.format(spec))

		return cls(**params)

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the logger.

			# Return value

			A lower-case string unique to this logger.
		"""
		return cls.__name__.lower()

	###########################################################################
	@staticmethod
	def get_all_loggers():
		""" Returns an iterator to the names of all logger.
		"""
		for cls in get_subclasses(Logger):
			yield cls

	###########################################################################
	@staticmethod
	def get_logger_by_name(name):
		""" Finds a logger class with the given name.
		"""
		name = name.lower()
		for cls in Logger.get_all_loggers():
			if cls.get_name() == name:
				return cls
		raise ValueError('No such logger with name "{}"'.format(name))

	###########################################################################
	def __init__(self, keep_batch=False, rate=None, **kwargs):
		""" Creates a new logger.

			# Arguments

			keep_batch: bool (default: False). If True, batch-level statistics
				are kept. Otherwise, they are silently discarded. Setting this
				to False can help with logging performance.
			rate: int or None (default: None). If `keep_batch` is True, then
				batch-level statistics will not be written more often than
				once every `rate` seconds (unless an epoch is completed, in
				which case batch statistics are written regardless of the value
				of `rate`, assuming `keep_batch` is True). If `rate` is zero,
				then batch information is written as quickly as possible (every
				batch). If `rate` is None, then, batch information is written
				only every epoch (equivalent to an arbitrarily large value of
				`rate`).
		"""
		if kwargs:
			logger.warning('Unexpected or unsupported arguments to the logger: '
			'%s', ', '.join(str(k) for k in kwargs.keys()))

		self.keep_batch = keep_batch
		if rate is not None:
			if not keep_batch:
				logger.warning('Logger rate is only meaningful when keep_batch '
					'is True. Ignoring the rate value.')
				rate = None
			if not isinstance(rate, int):
				raise ValueError('Logger rate must be None or an integer. '
					'Instead, we got: {}'.format(rate))
		self.rate = rate

		self.data = None
		self.timer = Timer(started=False)

		self._clear()

	###########################################################################
	def _clear(self):
		""" Clears the stored data.
		"""
		self.data = {k : {} for k in ('batch', 'training', 'validation')}

	###########################################################################
	@staticmethod
	def _get_dtype(entry):
		""" Finds an appropriate numpy datatype for the given data value.
		"""
		if isinstance(entry, numpy.ndarray):
			return entry.dtype
		elif isinstance(entry, (list, tuple)):
			if entry:
				return Logger._get_dtype(entry[0])
		elif isinstance(entry, int):
			return numpy.dtype('int32')

		# Floats, empty lists, and "other" types all get mapped to this default.
		return numpy.dtype('float32')

	###########################################################################
	@staticmethod
	def _get_shape(entry):
		""" Determines the shape of a data value.
		"""
		if isinstance(entry, numpy.ndarray):
			return entry.shape
		elif isinstance(entry, (list, tuple)):
			if entry:
				return (len(entry), ) + Logger._get_shape(entry[0])

		return ()

	###########################################################################
	@staticmethod
	def _prepare_like(entry, num_entries):
		""" Allocates a numpy array to hold a certain number of data values.
		"""
		return numpy.empty(
			shape=(num_entries, ) + Logger._get_shape(entry),
			dtype=Logger._get_dtype(entry)
		)

	###########################################################################
	@staticmethod
	def _arrange(data):
		""" Neatly arranges data in a Numpy array.

			# Arguments

			data: deque. Each entry in the deque is a dictionary whose keys name
				the training statistic stored in the respective dictionary
				values.

			# Return value

			Dictionary with the same keys as the entries in `data`, but whose
			values are each an array containing the concatenated values from
			the respective keys in `data`.
		"""
		num_entries = len(data)
		result = None
		for i, entry in enumerate(data):
			if result is None:
				result = {k : Logger._prepare_like(v, num_entries)
					for k, v in entry.items()}
			for k, v in entry.items():
				result[k][i] = v
		return result

	###########################################################################
	def flush(self):
		""" Hook for asking the logger to process log information in its queue.
		"""
		with CriticalSection():
			for key, tags in self.data.items():
				for tag, data in tags.items():
					if len(data) == 0:
						continue
					data = self._arrange(data)
					self.process(data, key, tag)

			self._clear()
			self.timer.restart()

	###########################################################################
	def _push(self, data_type, tag, data):
		""" Helper function for pushing data into the storage queue.
		"""
		if tag not in self.data[data_type]:
			self.data[data_type][tag] = deque()
		if tag == 'loss':
			if 'total' not in data:
				data['total'] = sum(data.values())
		self.data[data_type][tag].append(data)

	###########################################################################
	def log_batch(self, data, tag=None):
		""" Log training information after a batch.
		"""
		if not self.keep_batch:
			return

		self._push('batch', tag, data)

		if not self.timer.started or \
				(self.rate is not None and self.timer.get() > self.rate):
			self.flush()

	###########################################################################
	def log_training(self, data, tag=None):
		""" Log training statistics after an epoch.
		"""
		self._push('training', tag, data)
		self.flush()

	###########################################################################
	def log_validation(self, data, tag=None):
		""" Log training statistics after a validation run.
		"""
		self._push('validation', tag, data)
		self.flush()

	###########################################################################
	def get_best_training_loss(self):
		""" Returns the best historical training loss.

			# Return value

			If historical training loss is present in the log, it is returned.
			Otherwise, None is returned.
		"""
		raise NotImplementedError

	###########################################################################
	def get_best_validation_loss(self):
		""" Returns the best historical validation loss.

			# Return value

			If historical validation loss is present in the log, it is returned.
			Otherwise, None is returned.
		"""
		raise NotImplementedError

	###########################################################################
	def get_number_of_epochs(self):
		""" Returns the number of epochs this model has historically completed.

			# Return value

			If the number of epochs is known from the logs, it is returned.
			Otherwise, None is returned.
		"""
		raise NotImplementedError

	###########################################################################
	def process(self, data, data_type, tag=None):
		""" Processes training statistics.

			This should be implemented in derived classes, and provides the
			primary point for derived classes to receive data. This layer of
			indirection between the `log_X()` functions and this `process()`
			function is an optimization that improves performance by throttling
			the rate at which log data is processed.

			# Arguments

			data: dict. The training statistics dictionary, whose keys name the
				particular statistic, and the values are numpy arrays containing
				the most recent log data.
			data_type: str (one of: "batch", "training", "validation").
				Indicates what stage of training this statistics represent.
			tag: str or None (default: None). A tag to describe the type of data
				provided by `data` (e.g., "loss" or "accuracy").

			# Return value

			None

			# Notes

			- Although the `log_X()` functions accept dictionary values which
			  may be raw floats, `process()` only ever deals with values that
			  are arrays. The base class will collate the information into these
			  arrays so that they can be efficiently parsed by the child
			  classes.
		"""
		raise NotImplementedError

# Need to put this important here to avoid circular dependencies.
from .binary_logger import BinaryLogger	# pylint: disable=wrong-import-position
Logger.DEFAULT_LOGGER = BinaryLogger

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import numpy

from . import Provider
from ..sources import Shuffleable

###############################################################################
class ShuffleProvider(Provider): \
	# pylint: disable=too-few-public-methods,abstract-method
	""" Data provider that can shuffle data in-sync with support data sources.

		# Expected behavior

		This extends the list of expected behaviors from the `Provider` base
		class.

		- Because Providers create new Source iterators for each epoch, and
		  because each epoch ends whenever any Source iterator terminates (that
		  is, once the shorted dataset terminates), Providers only shuffle
		  Sources once every epoch, and can do so at the beginning of each
		  epoch. There is no need for Sources to indicate when they would like
		  to be shuffled.
		- There are two possibilities when applying shuffles:

			1. Past shuffles that have been applied to Sources are persistent,
			   affecting the Source's data in a non-reverseable way (as far as
			   the Source is concerned).
			2. Shuffles are temporary, and once the Source iterator is
			   recreated (between epochs), it can be assumed that the data---
			   prior to another shuffle---is in the same, original ordering.

		  These distinctions are transparent to the provider, which merely
		  needs to provide new shuffle indices. But it is possible for the
		  provider to help orchestrate these (e.g.. by providing reverse
		  permutations after each epoch).

		  Now, all Sources must be Shuffleable in order to use shuffling. This
		  means that even mixed finite/infinite Sources must support the
		  Shuffleable protocol. Sources that stream data samples (e.g.,
		  real-time data augmentation generators) could possibly be simplified
		  if the shuffle is assumed to be "new" after each epoch. Indeed, there
		  may be performance benefits (cache-coherency) from making certain
		  assumptions about how the shuffle "ought" to be applied.

		  Despite this, it would be imprudent to decide that Shuffleable
		  sources must operate under any particular set of assumptions about
		  how they store or organize data. Therefore, we opt for (1) and assume
		  that shuffles are persistent, and leave it to the individual sources
		  to decide how to combine multiple shuffles.
	"""

	###########################################################################
	def __init__(self, *args, randomize=True, **kwargs):
		""" Create a new data provider that can shuffle Shuffleable sources.

			# Arguments

			sources: dict or list. If this is a list, then it is a list of
				Sources. If this is a dictionary, its values are Sources and
				its keys are string names that are used by the training process
				to determine which nodes in the network should receive which
				information.
		"""
		super().__init__(*args, **kwargs)

		if randomize:
			x = [isinstance(source, Shuffleable) for source in self.sources]
			if not all(x):
				raise ValueError('All data sources must be Shuffleable for '
					'the provider to able to shuffle them.')

			if randomize is True:
				if self.entries is None:
					raise ValueError('Must know how long our shuffleable '
						'sources are in order to shuffle them, but all '
						'sources seem to be infinite. If this is the case, '
						'then set `randomize` to an integer in the '
						'Provider\'s constructor, or disable shuffling '
						'entirely by setting `randomize` to False.')
				self._shuffle_len = self.entries
			elif isinstance(randomize, int):
				self._shuffle_len = randomize
			else:
				raise ValueError('`randomize` must be True/False or an '
					'integer, but we received: {}'.format(randomize))

			self.randomize = True
		else:
			self.randomize = False

	###########################################################################
	def pre_iter(self):
		""" Pre-iteration hook.

			This is our opportunity to shuffle the data sources.
		"""
		super().pre_iter()

		if self.randomize:
			indices = numpy.random.permutation(self._shuffle_len)
			for source in self.sources:
				source.shuffle(indices)

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from .subclass import get_subclasses
from .flatiter import flatten, concatenate
from .iterhelp import get_any_value, merge_dict
from .importhelp import can_import
from .network import download_file
from .timer import Timer
from .critical import CriticalSection
from . import idx
from . import mergetools
from .environtools import EnvironmentalVariable
from .contexttools import redirect_stderr

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from . import Layer, ParsingError

###############################################################################
class Expand(Layer):					# pylint: disable=too-few-public-methods
	""" An expand layer adds a dimension of size 1.

		# Usage

		```
		expand:
		  dimension: 0
		```

		`DIM` must be an integer. It can be negative to count from the end.
	"""

	###########################################################################
	def __init__(self, *args, **kwargs):
		""" Creates a new expand layer.
		"""
		super().__init__(*args, **kwargs)
		self.dimension = None

	###########################################################################
	def _parse(self, engine):
		""" Parses out the expand layer.
		"""

		# Always call the parent.
		super()._parse(engine)

		if isinstance(self.args, dict) and 'dimension' in self.args:
			self.dimension = engine.evaluate(self.args['dimension'])
		elif isinstance(self.args, (str, int)):
			self.dimension = self.args
		else:
			raise ParsingError('The arguments to the "expand" layer must be an '
				'integer, or dictionary with a "dimension" key and an integer '
				'value.')
		try:
			self.dimension = int(self.dimension)
		except ValueError:
			raise ParsingError('"dimension" must evaluate to an integer.')

	###########################################################################
	def _build(self, backend):
		""" Instantiates the layer with the given backend.
		"""
		if backend.get_name() == 'keras':

			import keras.layers as L			# pylint: disable=import-error
			import keras.backend as K			# pylint: disable=import-error

			target_dim = self.dimension
			if target_dim >= 0:
				target_dim += 1

			def expand_shape(input_shape):
				""" Computes the expanded shape.
				"""
				dim = target_dim
				if dim < 0:
					dim += len(input_shape) + 1
				return input_shape[:dim] + (1,) + input_shape[dim:]

			yield L.Lambda(
				lambda x: K.expand_dims(x, dim=target_dim),
				expand_shape,
				name=self.name
			)

		else:
			raise ValueError('Unknown or unsupported backend: {}'.format(backend))

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import yaml
from . import Reader

###############################################################################
class YamlReader(Reader):
	""" A YAML reader.
	"""

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the reader.
		"""
		return 'yaml'

	###########################################################################
	@classmethod
	def supported_filetypes(cls):
		""" Returns a list of supported file extensions.
		"""
		return ('yml', 'yaml')

	###########################################################################
	def read(self, data):						# pylint: disable=no-self-use
		""" Reads the data and returns a native Python dictionary.

			# Arguments

			data: str. The data string to parse.

			# Return value

			A Python dictionary or list representing the data.

			# Exceptions

			If there is a YAML syntax/parsing error, a `yaml.error.YAMLError`
			instance is raised.
		"""
		return yaml.load(data)

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

from collections import OrderedDict

###############################################################################
def get_any_value(x):
	""" Returns a value of a dict, list, or tuple.

		# Arguments

		x: object. The object to grab a value for.

		# Return value

		If `x` is a dictionary (including an OrderedDict), returns a value from
		it. If `x` is a list or tuple, returns an element from it. Otherwise,
		raises ValueError.
	"""
	if isinstance(x, (dict, OrderedDict)):
		for v in x.values():
			return v
	if isinstance(x, (list, tuple)):
		return x[0]
	raise ValueError('Unexpected data type: {}'.format(type(x)))

###############################################################################
def get_any_key(x):
	""" Returns a key from a dictionary.

		# Arguments:

		x: dict. The dictionary to pull a key from.

		# Return value

		If x is empty, returns None; otherwise, this returns a key from the
		dictionary.
	"""
	if x:
		for k in x:
			return k
	else:
		return None

###############################################################################
def merge_dict(*args):
	""" Merges any number of dictionaries into a single dictionary.

		# Notes

		In Python 3.5+, you can just do this:
		```python
		r = {**x, **y}
		```
		But if you want a single expression in Python 3.4 and below:
		```python
		r = merge_dict(x, y)
		```
	"""
	result = {}
	for x in args:
		result.update(x)
	return result

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import logging
import numpy
from . import EvaluationHook

logger = logging.getLogger(__name__)

###############################################################################
class MnistHook(EvaluationHook):
	""" Post-evaluation hook for MNIST data, which prints summary statistics
		specific to the MNIST data set.
	"""

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the evaluation hook.
		"""
		return 'mnist'

	###########################################################################
	def apply(self, data, truth=None):
		""" Applies the hook to the data.
		"""
		if truth is None:
			logger.warning('MNIST hook only works with ground-truth data. We '
				'will skip this hook.')
			return data

		stats = {
			i : {'correct' : 0, 'total' : 0}
			for i in range(10)
		}

		if len(data) == 1:
			for key in data:
				pass
		elif len(data) > 1:
			if 'labels' in data:
				key = 'labels'
			else:
				logger.error('Cannot parse this data in the MNIST hook. Too '
					'many keys, none of which is "labels".')
				return data
		else:
			logger.error('MNIST hook cannot process this output: it has no '
				'keys.')
			return data

		if key not in truth:
			logger.error('No ground truth data present for MNIST hook.')
			return data

		if len(truth[key]) != len(data[key]):
			logger.warning('There appears to be a different amount of ground '
				'truth information than model output.')
			return data

		decoded_estimate = numpy.argmax(data[key], axis=1)
		decoded_truth = numpy.argmax(truth[key], axis=1)

		for estimate, truth in zip(decoded_estimate, decoded_truth):
			if truth not in stats:
				logger.warning('Out-of-range truth information: %d', truth)
			else:
				stats[truth]['total'] += 1
				if truth == estimate:
					stats[truth]['correct'] += 1

		print(('{: <10s}' * 4).format(
			'LABEL', 'CORRECT', 'TOTAL', 'ACCURACY'
		))
		for i in range(10):
			print(('{: <10d}'*3 + '{: >5.1f}%').format(
				i, stats[i]['correct'], stats[i]['total'],
				stats[i]['correct'] * 100. / stats[i]['total'] \
					if stats[i]['total'] else 0
			))

		correct = sum(x['correct'] for x in stats.values())
		total = sum(x['total'] for x in stats.values())
		print(('{: <10s}' + '{: <10d}'*2 + '{: >5.1f}%').format(
			'ALL', correct, total, correct * 100. / total if total else 0
		))

		return data

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF



"""
Copyright 2016 Deepgram

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

   http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import pickle
import logging
from . import EvaluationHook

logger = logging.getLogger(__name__)

###############################################################################
class OutputHook(EvaluationHook):
	""" Evaluation hook for saving to disk.
	"""

	###########################################################################
	@classmethod
	def get_name(cls):
		""" Returns the name of the evaluation hook.
		"""
		return 'output'

	###########################################################################
	@staticmethod
	def _save_as_pickle(target, data, truth=None):
		""" Saves a file as a Python 3 pickle.
		"""
		logger.info('Saving model output as pickle: %s', target)
		result = {'result' : data}
		if truth is not None:
			result['truth'] = truth
		with open(target, 'wb') as fh:
			pickle.dump(result, fh)

	###########################################################################
	def __init__(self, path=None, format=None, **kwargs): \
		# pylint: disable=redefined-builtin
		""" Creates a new output hook.
		"""

		super().__init__(**kwargs)

		if path is None:
			raise ValueError('No path specified in output hook.')

		self.path = path

		format = format or 'pkl'
		savers = {
			'pkl' : OutputHook._save_as_pickle,
			'pickle' : OutputHook._save_as_pickle
		}

		self.saver = savers.get(format)
		if self.saver is None:
			raise ValueError('No such handler for file format: {}'.format(
				format))

	###########################################################################
	def apply(self, data, truth=None):
		""" Applies the hook to the data.
		"""
		self.saver(
			target=self.path,
			data=data,
			truth=truth
		)
		return data

### EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF.EOF
